#
# bugtask.py - Internal helper class for sponsor-patch
#
# Copyright (C) 2010-2011, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import re
from urllib.parse import unquote
from urllib.request import urlretrieve

import distro_info
import httplib2

from ubuntutools.version import Version

Logger = logging.getLogger(__name__)


def is_sync(bug):
    """Checks if a Launchpad bug is a sync request.

    Either the title contains the word sync or the bug contains the tag sync."""

    return "sync" in bug.title.lower().split(" ") or "sync" in bug.tags


class BugTask:
    def __init__(self, bug_task, launchpad):
        self.bug_task = bug_task
        self.launchpad = launchpad

        components = re.split(r" \(| ", self.bug_task.bug_target_name.strip(")"))
        assert len(components) >= 1 and len(components) <= 3
        if len(components) == 1:
            self.package = None
            self.project = components[0]
            self.series = None
        elif len(components) == 2:
            self.package = components[0]
            self.project = components[1].lower()
            self.series = None
        elif len(components) == 3:
            self.package = components[0]
            self.project = components[1].lower()
            self.series = components[2].lower()

        if self.package is None:
            title_re = r"^Sync ([a-z0-9+.-]+) [a-z0-9.+:~-]+ \([a-z]+\) from.*"
            match = re.match(title_re, self.get_bug_title(), re.U | re.I)
            if match is not None:
                self.package = match.group(1)

    def download_source(self):
        source_files = self.get_source().sourceFileUrls()
        dsc_file = ""
        for url in source_files:
            filename = unquote(os.path.basename(url))
            Logger.debug("Downloading %s...", filename)
            # HttpLib2 isn't suitable for large files (it reads into memory),
            # but we want its https certificate validation on the .dsc
            if url.endswith(".dsc"):
                response, data = httplib2.Http().request(url)
                assert response.status == 200
                with open(filename, "wb") as f:
                    f.write(data)

                dsc_file = os.path.join(os.getcwd(), filename)
            else:
                urlretrieve(url, filename)
        assert os.path.isfile(dsc_file), f"{dsc_file} does not exist."
        return dsc_file

    def get_branch_link(self):
        return "lp:" + self.project + "/" + self.get_series() + "/" + self.package

    def get_bug_title(self):
        """Returns the title of the related bug."""
        return self.bug_task.bug.title

    def get_long_info(self):
        return (
            "Bug task: "
            + str(self.bug_task)
            + "\n"
            + "Package: "
            + str(self.package)
            + "\n"
            + "Project: "
            + str(self.project)
            + "\n"
            + "Series: "
            + str(self.series)
        )

    def get_lp_task(self):
        """Returns the Launchpad bug task object."""
        return self.bug_task

    def get_package_and_series(self):
        result = self.package
        if self.series:
            result += " (" + self.series + ")"
        return result

    def get_previous_version(self):
        if self.is_derived_from_debian():
            previous_version = self.get_latest_released_version()
        else:
            previous_version = self.get_version()
        return previous_version

    def get_series(self, latest_release=False):
        if self.series is None or latest_release:
            dist = self.launchpad.distributions[self.project]
            return dist.current_series.name
        return self.series

    def get_short_info(self):
        return self.bug_task.bug_target_name + ": " + self.bug_task.status

    def get_source(self, latest_release=False):
        assert self.package is not None

        if self.is_derived_from_debian() and not latest_release:
            project = "debian"
            series = self.get_debian_source_series()
        else:
            project = self.project
            series = self.get_series(latest_release)

        dist = self.launchpad.distributions[project]
        archive = dist.getArchive(name="primary")
        distro_series = dist.getSeries(name_or_version=series)
        published = archive.getPublishedSources(
            source_name=self.package,
            distro_series=distro_series,
            status="Published",
            exact_match=True,
        )

        latest_source = None
        for source in published:
            if source.pocket in ("Release", "Security", "Updates", "Proposed"):
                latest_source = source
                break
        return latest_source

    def get_version(self):
        source_package_version = self.get_source().source_package_version
        return Version(source_package_version)

    def get_latest_released_version(self):
        source = self.get_source(True)
        if source is None:  # Not currently published in Ubuntu
            version = "~"
        else:
            version = source.source_package_version
        return Version(version)

    def get_debian_source_series(self):
        title = self.bug_task.bug.title.lower().split()
        if "experimental" in title:
            series = "experimental"
        elif "testing" in title:
            series = distro_info.DebianDistroInfo().testing()
        else:
            series = distro_info.DebianDistroInfo().devel()
        return series

    def is_complete(self):
        return self.bug_task.is_complete

    def is_derived_from_debian(self):
        """Checks if this task get's the source from Debian."""
        return self.is_merge() or self.is_sync()

    def is_merge(self):
        bug = self.bug_task.bug
        return "merge" in bug.title.lower().split(" ") or "merge" in bug.tags

    def is_sync(self):
        return is_sync(self.bug_task.bug)

    def is_ubuntu_task(self):
        return self.project == "ubuntu"

    def title_contains(self, word):
        """Checks if the bug title contains the given word."""
        return word in self.bug_task.bug.title.split(" ")



# example_package.py - Creates an example package
#
# Copyright (C) 2010-2011, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import os
import subprocess
import tempfile
from pathlib import Path

from ubuntutools.version import Version


class ExamplePackage:
    def __init__(self, source="example", version="1.0-1", destdir="test-data"):
        self.source = source
        self.version = Version(version)
        self.destdir = Path(destdir)

        self.env = dict(os.environ)
        self.env["DEBFULLNAME"] = "Example"
        self.env["DEBEMAIL"] = "example@example.net"

    @property
    def orig(self):
        return self.destdir / f"{self.source}_{self.version.upstream_version}.orig.tar.xz"

    @property
    def debian(self):
        return self.destdir / f"{self.source}_{self.version}.debian.tar.xz"

    @property
    def dsc(self):
        return self.destdir / f"{self.source}_{self.version}.dsc"

    @property
    def dirname(self):
        return f"{self.source}-{self.version.upstream_version}"

    @property
    def content_filename(self):
        return "content"

    @property
    def content_text(self):
        return "my content"

    def create(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            self._create(Path(tmpdir))

    def _create(self, directory: Path) -> None:
        pkgdir = directory / self.dirname
        pkgdir.mkdir()
        (pkgdir / self.content_filename).write_text(self.content_text)

        # run dh_make to create orig tarball
        subprocess.run(
            "dh_make -sy --createorig".split(),
            check=True,
            env=self.env,
            cwd=str(pkgdir),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        # run dpkg-source -b to create debian tar and dsc
        subprocess.run(
            f"dpkg-source -b {self.dirname}".split(),
            check=True,
            env=self.env,
            cwd=str(directory),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        # move tarballs and dsc to destdir
        self.destdir.mkdir(parents=True, exist_ok=True)
        (directory / self.orig.name).rename(self.orig)
        (directory / self.debian.name).rename(self.debian)
        (directory / self.dsc.name).rename(self.dsc)



#
# patch.py - Internal helper class for sponsor-patch
#
# Copyright (C) 2010-2011, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import re
import subprocess
from functools import reduce

from ubuntutools.sponsor_patch.question import ask_for_manual_fixing

Logger = logging.getLogger(__name__)


class Patch:
    """This object represents a patch that can be downloaded from Launchpad."""

    def __init__(self, patch):
        self._patch = patch
        self._patch_file = re.sub(" |/", "_", patch.title)
        if not reduce(
            lambda r, x: r or self._patch.title.endswith(x), (".debdiff", ".diff", ".patch"), False
        ):
            Logger.debug("Patch %s does not have a proper file extension.", self._patch.title)
            self._patch_file += ".patch"
        self._full_path = os.path.realpath(self._patch_file)
        self._changed_files = None

    def apply(self, task):
        """Applies the patch in the current directory."""
        assert self._changed_files is not None, "You forgot to download the patch."
        edit = False
        if self.is_debdiff():
            cmd = [
                "patch",
                "--merge",
                "--force",
                "-p",
                str(self.get_strip_level()),
                "-i",
                self._full_path,
            ]
            Logger.debug(" ".join(cmd))
            if subprocess.call(cmd) != 0:
                Logger.error(
                    "Failed to apply debdiff %s to %s %s.",
                    self._patch_file,
                    task.package,
                    task.get_version(),
                )
                if not edit:
                    ask_for_manual_fixing()
                    edit = True
        else:
            cmd = ["add-patch", self._full_path]
            Logger.debug(" ".join(cmd))
            if subprocess.call(cmd) != 0:
                Logger.error(
                    "Failed to apply diff %s to %s %s.",
                    self._patch_file,
                    task.package,
                    task.get_version(),
                )
                if not edit:
                    ask_for_manual_fixing()
                    edit = True
        return edit

    def download(self):
        """Downloads the patch from Launchpad."""
        Logger.debug("Downloading %s.", self._patch_file)
        patch_f = open(self._patch_file, "wb")
        patch_f.write(self._patch.data.open().read())
        patch_f.close()

        cmd = ["diffstat", "-l", "-p0", self._full_path]
        changed_files = subprocess.check_output(cmd, encoding="utf-8")
        self._changed_files = [f for f in changed_files.split("\n") if f != ""]

    def get_strip_level(self):
        """Returns the stript level for the patch."""
        assert self._changed_files is not None, "You forgot to download the patch."
        strip_level = None
        if self.is_debdiff():
            changelog = [f for f in self._changed_files if f.endswith("debian/changelog")][0]
            strip_level = len(changelog.split(os.sep)) - 2
        return strip_level

    def is_debdiff(self):
        """Checks if the patch is a debdiff (= modifies debian/changelog)."""
        assert self._changed_files is not None, "You forgot to download the patch."
        return len([f for f in self._changed_files if f.endswith("debian/changelog")]) > 0



# -*- coding: utf-8 -*-
#
#   lp.py - methods used by requestsync while interacting
#           directly with Launchpad
#
#   Copyright © 2009 Michael Bienia <geser@ubuntu.com>
#
#   This module may contain code written by other authors/contributors to
#   the main requestsync script. See there for their names.
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; version 2
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   Please see the /usr/share/common-licenses/GPL-2 file for the full text
#   of the GNU General Public License license.

import logging
import re

from debian.deb822 import Changes
from distro_info import DebianDistroInfo, DistroDataOutdated
from httplib2 import Http, HttpLib2Error

from ubuntutools.lp import udtexceptions
from ubuntutools.lp.lpapicache import (
    Distribution,
    DistributionSourcePackage,
    Launchpad,
    PersonTeam,
)
from ubuntutools.question import confirmation_prompt

Logger = logging.getLogger(__name__)


def get_debian_srcpkg(name, release):
    debian = Distribution("debian")
    debian_archive = debian.getArchive()

    try:
        release = DebianDistroInfo().codename(release, None, release)
    except DistroDataOutdated as e:
        Logger.warning(e)

    return debian_archive.getSourcePackage(name, release)


def get_ubuntu_srcpkg(name, release, pocket="Release"):
    ubuntu = Distribution("ubuntu")
    ubuntu_archive = ubuntu.getArchive()

    try:
        return ubuntu_archive.getSourcePackage(name, release, pocket)
    except udtexceptions.PackageNotFoundException:
        if pocket != "Release":
            parent_pocket = "Release"
            if pocket == "Updates":
                parent_pocket = "Proposed"
            return get_ubuntu_srcpkg(name, release, parent_pocket)
        raise


def need_sponsorship(name, component, release):
    """
    Check if the user has upload permissions for either the package
    itself or the component
    """
    archive = Distribution("ubuntu").getArchive()
    distroseries = Distribution("ubuntu").getSeries(release)

    need_sponsor = not PersonTeam.me.canUploadPackage(archive, distroseries, name, component)
    if need_sponsor:
        print("""You are not able to upload this package directly to Ubuntu.
Your sync request shall require an approval by a member of the appropriate
sponsorship team, who shall be subscribed to this bug report.
This must be done before it can be processed by a member of the Ubuntu Archive
team.""")
        confirmation_prompt()

    return need_sponsor


def check_existing_reports(srcpkg):
    """
    Check existing bug reports on Launchpad for a possible sync request.

    If found ask for confirmation on filing a request.
    """

    # Fetch the package's bug list from Launchpad
    pkg = Distribution("ubuntu").getSourcePackage(name=srcpkg)
    pkg_bug_list = pkg.searchTasks(
        status=["Incomplete", "New", "Confirmed", "Triaged", "In Progress", "Fix Committed"],
        omit_duplicates=True,
    )

    # Search bug list for other sync requests.
    for bug in pkg_bug_list:
        # check for Sync or sync and the package name
        if not bug.is_complete and f"ync {srcpkg}" in bug.title:
            print(
                f"The following bug could be a possible duplicate sync bug on Launchpad:\n"
                f" * {bug.title} ({bug.web_link})\n"
                f"Please check the above URL to verify this before continuing."
            )
            confirmation_prompt()


def get_ubuntu_delta_changelog(srcpkg):
    """
    Download the Ubuntu changelog and extract the entries since the last sync
    from Debian.
    """
    archive = Distribution("ubuntu").getArchive()
    spph = archive.getPublishedSources(
        source_name=srcpkg.getPackageName(), exact_match=True, pocket="Release"
    )
    debian_info = DebianDistroInfo()
    name_chars = "[-+0-9a-z.]"
    topline = re.compile(
        rf"^(\w%({name_chars})s*) \(([^\(\) \t]+)\)((\s+%({name_chars})s+)+)\;", re.IGNORECASE
    )
    delta = []
    for record in spph:
        changes_url = record.changesFileUrl()
        if changes_url is None:
            # Native sync
            break
        try:
            response = Http().request(changes_url)[0]
        except HttpLib2Error as e:
            Logger.error(str(e))
            break
        if response.status != 200:
            Logger.error("%s: %s %s", changes_url, response.status, response.reason)
            break

        changes = Changes(Http().request(changes_url)[1])
        for line in changes["Changes"].splitlines():
            line = line[1:]
            match = topline.match(line)
            if match:
                distribution = match.group(3).split()[0].split("-")[0]
                if debian_info.valid(distribution):
                    break
            if line.startswith("  "):
                delta.append(line)
        else:
            continue
        break

    return "\n".join(delta)


def post_bug(srcpkg, subscribe, status, bugtitle, bugtext):
    """
    Use the LP API to file the sync request.
    """

    print(f"The final report is:\nSummary: {bugtitle}\nDescription:\n{bugtext}\n")
    confirmation_prompt()

    if srcpkg:
        # pylint: disable=protected-access
        bug_target = DistributionSourcePackage(f"{Launchpad._root_uri}ubuntu/+source/{srcpkg}")
    else:
        # new source package
        bug_target = Distribution("ubuntu")

    # create bug
    bug = Launchpad.bugs.createBug(title=bugtitle, description=bugtext, target=bug_target())

    # newly created bugreports have only one task
    task = bug.bug_tasks[0]
    # only members of ubuntu-bugcontrol can set importance
    if PersonTeam.me.isLpTeamMember("ubuntu-bugcontrol"):
        task.importance = "Wishlist"
    task.status = status
    task.lp_save()

    bug.subscribe(person=PersonTeam(subscribe)())

    print(f"Sync request filed as bug #{bug.id}: {bug.web_link}")



class PackageNotFoundException(BaseException):
    """Thrown when a package is not found"""


class SeriesNotFoundException(BaseException):
    """Thrown when a distroseries is not found"""


class PocketDoesNotExistError(Exception):
    """Raised when a invalid pocket is used."""


class ArchiveNotFoundException(BaseException):
    """Thrown when an archive for a distibution is not found"""


class AlreadyLoggedInError(Exception):
    """Raised when a second login is attempted."""


class ArchSeriesNotFoundException(BaseException):
    """Thrown when a distroarchseries is not found."""


class InvalidDistroValueError(ValueError):
    """Thrown when distro value is invalid"""



# archive.py - Functions for dealing with Debian source packages, archives,
#              and mirrors.
#
# Copyright (C) 2010-2011, Stefano Rivera <stefanor@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

"""Pull source packages from archives.

Approach:
1. Pull dsc from Launchpad (this is over https and can authenticate the
   rest of the source package)
2. Attempt to pull the remaining files from:
   1. existing files
   2. mirrors
   3. Launchpad
3. Verify checksums.
"""

import codecs
import functools
import json
import logging
import os
import re
import subprocess
import sys
import tempfile
from abc import ABC, abstractmethod
from contextlib import closing
from pathlib import Path
from urllib.parse import urljoin, urlparse
from urllib.request import urlopen

import debian.deb822
from debian.changelog import Changelog

from ubuntutools.config import UDTConfig
from ubuntutools.lp.lpapicache import (
    Distribution,
    HTTPError,
    Launchpad,
    PersonTeam,
    Project,
    SourcePackagePublishingHistory,
)
from ubuntutools.lp.udtexceptions import (
    InvalidDistroValueError,
    PackageNotFoundException,
    PocketDoesNotExistError,
    SeriesNotFoundException,
)
from ubuntutools.misc import (
    DownloadError,
    NotFoundError,
    download,
    download_bytes,
    verify_file_checksum,
    verify_file_checksums,
)
from ubuntutools.version import Version

Logger = logging.getLogger(__name__)


class Dsc(debian.deb822.Dsc):
    "Extend deb822's Dsc with checksum verification abilities"

    def get_strongest_checksum(self):
        "Return alg, dict by filename of size, hash_ pairs"
        if "Checksums-Sha256" in self:
            return (
                "sha256",
                dict(
                    (entry["name"], (int(entry["size"]), entry["sha256"]))
                    for entry in self["Checksums-Sha256"]
                ),
            )
        if "Checksums-Sha1" in self:
            return (
                "sha1",
                dict(
                    (entry["name"], (int(entry["size"]), entry["sha1"]))
                    for entry in self["Checksums-Sha1"]
                ),
            )
        return (
            "md5",
            dict(
                (entry["name"], (int(entry["size"]), entry["md5sum"])) for entry in self["Files"]
            ),
        )

    def verify_file(self, pathname):
        "Verify that pathname matches the checksums in the dsc"
        path = Path(pathname)
        if not path.is_file():
            return False
        alg, checksums = self.get_strongest_checksum()
        size, digest = checksums[path.name]
        return verify_file_checksum(path, alg, digest, size)

    def compare_dsc(self, other):
        """Check whether any files in these two dscs that have the same name
        also have the same checksum."""
        for field, key in (
            ("Checksums-Sha256", "sha256"),
            ("Checksums-Sha1", "sha1"),
            ("Files", "md5sum"),
        ):
            if field not in self or field not in other:
                continue
            our_checksums = dict(
                (entry["name"], (int(entry["size"]), entry[key])) for entry in self[field]
            )
            their_checksums = dict(
                (entry["name"], (int(entry["size"]), entry[key])) for entry in other[field]
            )
            for name, (size, checksum) in our_checksums.items():
                if name not in their_checksums:
                    # file only in one dsc
                    continue
                if size != their_checksums[name][0] or checksum != their_checksums[name][1]:
                    return False
            return True  # one checksum is good enough
        return True


class SourcePackage(ABC):
    """Base class for source package downloading.
    Use DebianSourcePackage or UbuntuSourcePackage instead of using this
    directly.
    """

    @property
    @abstractmethod
    def distribution(self):
        return None

    @property
    def spph_class(self):
        return SourcePackagePublishingHistory

    def __init__(self, package=None, version=None, component=None, **kwargs):
        """Can be initialised using either package or dscfile.
        If package is specified, either the version or series can also be
        specified; using version will get the specific package version,
        while using the series will get the latest version from that series.
        Specifying only the package with no version or series will get the
        latest version from the development series.
        """
        dscfile = kwargs.get("dscfile")
        mirrors = kwargs.get("mirrors", ())
        workdir = kwargs.get("workdir")
        series = kwargs.get("series")
        pocket = kwargs.get("pocket")
        status = kwargs.get("status")
        arch = kwargs.get("arch")
        verify_signature = kwargs.get("verify_signature", False)
        try_binary = kwargs.get("try_binary", True)

        assert package is not None or dscfile is not None

        if "lp" in kwargs:
            # deprecated - please don't use this; if you want to use an
            # existing lp object, just call login_existing() directly
            Logger.warning("Deprecation warning: please don't pass 'lp' to SourcePackage")
            if not Launchpad.logged_in:
                Launchpad.login_existing(kwargs["lp"])

        self.source = package
        self.binary = None
        self.try_binary = try_binary
        self.workdir = Path(workdir) if workdir else Path(".")
        self._series = series
        self._pocket = pocket
        self._status = status
        self._arch = arch
        # dscfile can be either a path or an URL.  misc.py's download() will
        # later fiture it out
        self._dsc_source = dscfile
        self._verify_signature = verify_signature

        # Cached values:
        self._distribution = None
        self._component = component
        self._dsc = None
        self._spph = None
        self._version = Version(version) if version else None

        # Mirrors
        self.mirrors = list(filter(None, mirrors))
        self.masters = list(
            filter(
                None,
                [
                    UDTConfig.defaults.get(f"{self.distribution.upper()}_{suffix}")
                    for suffix in ["MIRROR", "PORTS_MIRROR", "INTERNAL_MIRROR"]
                ],
            )
        )

        # If provided a dscfile, process it now to set our source and version
        if self._dsc_source:
            self._dsc = Dsc(download_bytes(self._dsc_source))
            self.source = self._dsc["Source"]
            self._version = Version(self._dsc["Version"])
            self._check_dsc_signature()

    @property
    def lp_spph(self):
        "Return the LP Source Package Publishing History entry"
        if self._spph:
            return self._spph

        archive = self.getArchive()
        params = {}
        if self._version:
            # if version was specified, use that
            params["version"] = self._version.full_version
        elif self._series:
            # if version not specified, get the latest from this series
            params["series"] = self._series
            # note that if not specified, pocket defaults to all EXCEPT -backports
            if self._pocket:
                params["pocket"] = self._pocket
        else:
            # We always want to search all series, if not specified
            params["search_all_series"] = True

        params["status"] = self._status

        try:
            self._spph = archive.getSourcePackage(self.source, wrapper=self.spph_class, **params)
            return self._spph
        except PackageNotFoundException as pnfe:
            if not self.try_binary or self.binary:
                # either we don't need to bother trying binary name lookup,
                # or we've already tried
                raise pnfe

            Logger.info(
                "Source package lookup failed, trying lookup of binary package %s", self.source
            )

            try:
                params["archtag"] = self._arch
                bpph = archive.getBinaryPackage(self.source, **params)
            except PackageNotFoundException as bpnfe:
                # log binary lookup failure, in case it provides hints
                Logger.info(str(bpnfe))
                # raise the original exception for the source lookup
                raise pnfe from None

            self.binary = self.source
            self.source = bpph.getSourcePackageName()
            Logger.info(
                "Using source package '%s' for binary package '%s'", self.source, self.binary
            )

            spph = bpph.getBuild().getSourcePackagePublishingHistory()
            if spph:
                self._spph = self.spph_class(spph.self_link)
                return self._spph

            # binary build didn't include source link, unfortunately
            # so try again with the updated self.source name
            if not self._version:
                # Get version first if user didn't specify it, as some
                # binaries have their version hardcoded in their name,
                # such as the kernel package
                self._version = Version(bpph.getVersion())
            return self.lp_spph

    @property
    def version(self):
        "Return Package version"
        if not self._version:
            self._version = Version(self.lp_spph.getVersion())
        return self._version

    @property
    def component(self):
        "Cached archive component, in available"
        if not self._component:
            Logger.debug("Determining component from Launchpad")
            self._component = self.lp_spph.getComponent()
        return self._component

    @property
    def dsc_name(self):
        "Return the source package dsc filename for the given package"
        return f"{self.source}_{self.version.strip_epoch()}.dsc"

    @property
    def dsc_pathname(self):
        "Return the dsc_name, with the workdir path"
        return str(self.workdir / self.dsc_name)

    @property
    def dsc(self):
        "Return the Dsc"
        if not self._dsc:
            if self._dsc_source:
                raise RuntimeError("Internal error: we have a dsc file but dsc not set")
            urls = self._source_urls(self.dsc_name)
            with tempfile.TemporaryDirectory() as tmpdir:
                tmpdsc = Path(tmpdir) / self.dsc_name
                self._download_file_from_urls(urls, tmpdsc)
                self._dsc = Dsc(tmpdsc.read_bytes())
                self._check_dsc_signature()
        return self._dsc

    def getDistribution(self):
        if not self._distribution:
            self._distribution = Distribution(self.distribution)
        return self._distribution

    def getArchive(self):
        return self.getDistribution().getArchive()

    def _mirror_url(self, mirror, component, filename):
        "Build a source package URL on a mirror"
        if self.source.startswith("lib"):
            group = self.source[:4]
        else:
            group = self.source[0]
        return os.path.join(mirror, "pool", component, group, self.source, filename)

    def _archive_servers(self):
        "Generator for mirror and master servers"
        # Always provide the mirrors first
        yield from self.mirrors
        # Don't repeat servers that are in both mirrors and masters
        yield from set(self.masters) - set(self.mirrors)

    def _source_urls(self, name):
        "Generator of sources for name"
        if self._dsc_source:
            # we only take "" as file, as regardless if for some reason this
            # is a file:// url we still need to handle it with urljoin
            if urlparse(str(self._dsc_source)).scheme == "":
                yield str(Path(self._dsc_source).parent / name)
            else:
                yield urljoin(self._dsc_source, name)
        for server in self._archive_servers():
            yield self._mirror_url(server, self.component, name)
        if self.lp_spph.sourceFileUrl(name):
            yield self.lp_spph.sourceFileUrl(name)

    def _binary_urls(self, name, bpph):
        "Generator of URLs for name"
        for server in self._archive_servers():
            yield self._mirror_url(server, bpph.getComponent(), name)
        if bpph.binaryFileUrl(name):
            yield bpph.binaryFileUrl(name)
        if bpph.getUrl():
            yield bpph.getUrl()

    def _check_dsc_signature(self):
        "Check that the dsc signature matches what we are expecting"
        if not self._verify_signature:
            return
        try:
            gpg_info = self.dsc.get_gpg_info(
                (
                    "/usr/share/keyrings/debian-keyring.gpg",
                    "/usr/share/keyrings/debian-maintainers.gpg",
                )
            )
        except IOError:
            Logger.debug(
                "Signature on %s could not be verified, install debian-keyring", self.dsc_name
            )
            return
        if gpg_info.valid():
            if "GOODSIG" in gpg_info:
                Logger.info(
                    "Good signature by %s (0x%s)", gpg_info["GOODSIG"][1], gpg_info["GOODSIG"][0]
                )
            elif "VALIDSIG" in gpg_info:
                Logger.info("Valid signature by 0x%s", gpg_info["VALIDSIG"][0])
            else:
                Logger.info("Valid signature")
        elif "NO_PUBKEY" in gpg_info:
            Logger.warning("Public key not found, could not verify signature")
        elif "NODATA" in gpg_info:
            Logger.warning("Package is not signed")
        else:
            Logger.warning("Signature on %s could not be verified", self.dsc_name)

    def _verify_file(self, pathname, dscverify=False, sha1sum=None, sha256sum=None, size=0):
        path = Path(pathname)
        if not path.exists():
            return False
        if dscverify and not self.dsc.verify_file(path):
            return False
        checksums = {}
        if sha1sum:
            checksums["SHA1"] = sha1sum
        if sha256sum:
            checksums["SHA256"] = sha256sum
        if not verify_file_checksums(path, checksums, size):
            return False
        return True

    def _download_file(self, url, filename, size=0, dscverify=False, sha1sum=None, sha256sum=None):
        "Download url to filename; will be put in workdir unless filename is absolute path."
        if Path(filename).is_absolute():
            path = Path(filename).expanduser().resolve()
        else:
            path = self.workdir / filename

        can_verify = any((dscverify, sha1sum, sha256sum))
        if can_verify and self._verify_file(path, dscverify, sha1sum, sha256sum, size):
            Logger.info("Using existing file %s", path)
            return True

        download(url, path, size)

        return self._verify_file(path, dscverify, sha1sum, sha256sum, size)

    def _download_file_from_urls(
        self, urls, filename, size=0, dscverify=False, sha1sum=None, sha256sum=None
    ):
        "Try to download a file from a list of urls."
        for url in urls:
            try:
                if self._download_file(
                    url, filename, size, dscverify=dscverify, sha1sum=sha1sum, sha256sum=sha256sum
                ):
                    return
            except NotFoundError:
                # It's ok if the file isn't found, just try the next url
                Logger.debug("File not found at %s", url)
            except DownloadError as e:
                Logger.error("Download Error: %s", e)
        raise DownloadError(f"Failed to download {filename}")

    @staticmethod
    def pull_dsc():
        """DEPRECATED

        This method is badly named and historically has only 'pulled' the
        dsc into memory, not actually to a file. Since the other 'pull' methods
        actually 'pull' to real files, this method makes no sense; additionally
        there is no need for this method since we can 'pull' the dsc into
        memory when the .dsc property is accessed.

        This method no longer does anything at all and should not be used by
        anyone.
        """
        Logger.debug('Please fix your program: the "pull_dsc" method is deprecated')

    def pull(self):
        "Pull into workdir"
        Path(self.dsc_pathname).write_bytes(self.dsc.raw_text)
        for entry in self.dsc["Files"]:
            name = entry["name"]
            urls = self._source_urls(name)
            self._download_file_from_urls(urls, name, int(entry["size"]), dscverify=True)

    def pull_binaries(self, arch=None, name=None, ext=None):
        """Pull binary debs into workdir.
        If name is specified, only binary packages matching the regex are included.

        If ext is specified, only binary packages with that ext are included; for
        example to only download dbgsym ddebs, specify ext='ddeb'.

        If arch is not specified or is 'all', pull all archs.

        Returns the number of files downloaded.
        """
        Logger.debug("pull_binaries(arch=%s, name=%s, ext=%s)", arch, name, ext)

        if arch == "all":
            arch = None

        total = 0
        for bpph in self.lp_spph.getBinaries(arch=arch, name=name, ext=ext):
            fname = bpph.getFileName()
            fsha1 = bpph.binaryFileSha1(fname)
            fsha256 = bpph.binaryFileSha256(fname)
            fsize = bpph.binaryFileSize(fname)
            urls = self._binary_urls(fname, bpph)
            try:
                self._download_file_from_urls(urls, fname, fsize, sha1sum=fsha1, sha256sum=fsha256)
                total += 1
            except DownloadError as e:
                # log/print the error, but continue to get the rest of the files
                Logger.error(e)
        return total

    def verify(self):
        """Verify that the source package in workdir matches the dsc.
        Return boolean
        """
        return all(
            self.dsc.verify_file(self.workdir / entry["name"]) for entry in self.dsc["Files"]
        )

    def verify_orig(self):
        """Verify that the .orig files in workdir match the dsc.
        Return boolean
        """
        orig_re = re.compile(r".*\.orig(-[^.]+)?\.tar\.[^.]+$")
        return all(
            self.dsc.verify_file(self.workdir / entry["name"])
            for entry in self.dsc["Files"]
            if orig_re.match(entry["name"])
        )

    def unpack(self, destdir=None):
        "Unpack in workdir"
        cmd = ["dpkg-source", "-x", self.dsc_name]
        if destdir:
            cmd.append(destdir)
        Logger.debug(" ".join(cmd))
        result = subprocess.run(
            cmd,
            check=False,
            cwd=str(self.workdir),
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if result.returncode != 0:
            Logger.error("Source unpack failed.")
            Logger.debug(result.stdout)

    def debdiff(self, newpkg, diffstat=False):
        """Write a debdiff comparing this src pkg to a newer one.
        Optionally print diffstat.
        Return the debdiff filename.
        """
        cmd = ["debdiff", self.dsc_name, newpkg.dsc_name]
        difffn = f"{newpkg.dsc_name[:-3]}debdiff"
        Logger.debug("%s > %s", " ".join(cmd), difffn)
        with open(difffn, "w", encoding="utf-8") as f:
            if subprocess.call(cmd, stdout=f, cwd=str(self.workdir)) > 2:
                Logger.error("Debdiff failed.")
                sys.exit(1)
        if diffstat:
            cmd = ("diffstat", "-p1", difffn)
            Logger.debug(" ".join(cmd))
            if subprocess.call(cmd):
                Logger.error("diffstat failed.")
                sys.exit(1)
        return os.path.abspath(difffn)


class DebianSPPH(SourcePackagePublishingHistory):
    """SPPH with getBinaries() overridden,
    as LP doesn't have Debian binaries
    """

    resource_type = "source_package_publishing_history"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._srcpkg = None

    def getBinaries(self, arch=None, name=None, ext=None):
        if not self._srcpkg:
            Logger.info("Using Snapshot to find binary packages")
            self._srcpkg = Snapshot.getSourcePackage(
                self.getPackageName(), version=self.getVersion()
            )
        return self._srcpkg.getSPPH().getBinaries(arch=arch, name=name, ext=ext)


class DebianSourcePackage(SourcePackage):
    "Download / unpack a Debian source package"

    @property
    def distribution(self):
        return "debian"

    @property
    def spph_class(self):
        return DebianSPPH

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.masters.append(UDTConfig.defaults["DEBSEC_MIRROR"])

        # Cached values:
        self._snapshot_package = None
        self._snapshot_files = None
        # Don't bother searching in LP for debian binaries, they aren't there
        self.try_binary = False

        # Debian doesn't have 'pockets'
        if self._pocket:
            if self._pocket.lower() != "release":
                Logger.error("Debian does not use 'pockets', ignoring pocket '%s'", self._pocket)
            self._pocket = None

    # Overridden properties/methods:
    @property
    def lp_spph(self):
        "Return the LP Source Package Publishing History entry"
        if not self._spph:
            try:
                # superclass will set self._spph
                return super().lp_spph
            except PackageNotFoundException:
                pass
            except SeriesNotFoundException:
                pass

            Logger.info("Package not found in Launchpad, using Snapshot")
            self._spph = self.snapshot_package.getSPPH()
        return self._spph

    @property
    def component(self):
        "Cached archive component, in available"
        if not self._component:
            Logger.debug("Determining component from Snapshot")
            self._component = Snapshot.getComponent(self.source, self.version)
        return self._component

    def _source_urls(self, name):
        "Generator of sources for name"
        yield from super()._source_urls(name)
        if name in self.snapshot_files:
            yield self.snapshot_files[name]

    # Local methods:
    @property
    def snapshot_package(self):
        if not self._snapshot_package:
            if self._version or self._spph:
                # as .version uses .lpph, and our .lpph might use us,
                # only use .version if _version or _spph are set
                version = self.version.full_version
                srcpkg = Snapshot.getSourcePackage(self.source, version=version)
                if not srcpkg:
                    msg = f"Package {self.source} {version} not found"
                    raise PackageNotFoundException(msg)
                self._snapshot_package = srcpkg
            else:
                # we have neither version nor spph, so look up our version using madison
                Logger.info("Using madison to find latest version number")
                series = self._series
                params = {"series": series} if series else {}
                srcpkg = Madison(self.distribution).getSourcePackage(self.source, **params)
                if not srcpkg:
                    raise PackageNotFoundException(f"Package {self.source} not found")
                if self.source != srcpkg.name:
                    self.binary = self.source
                    self.source = srcpkg.name
                self._snapshot_package = srcpkg
        return self._snapshot_package

    @property
    def snapshot_files(self):
        if not self._snapshot_files:
            files = self.snapshot_package.getFiles()
            self._snapshot_files = {f.name: f.getUrl() for f in files}
        return self._snapshot_files


class UbuntuSourcePackage(SourcePackage):
    "Download / unpack an Ubuntu source package"

    @property
    def distribution(self):
        return "ubuntu"


class PersonalPackageArchiveSourcePackage(UbuntuSourcePackage):
    "Download / unpack an Ubuntu Personal Package Archive source package"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert "ppa" in kwargs
        ppa = kwargs["ppa"].split("/")
        if len(ppa) != 2:
            raise ValueError(f'Invalid PPA value "{kwargs["ppa"]}",' 'must be "<USER>/<PPA>"')
        self._teamname = ppa[0]
        self._ppaname = ppa[1]
        self.masters = []

    @property
    @functools.lru_cache()
    def team(self):
        try:
            return PersonTeam.fetch(self._teamname)
        except KeyError as error:
            raise ValueError(f"No user/team '{self._teamname}' found on Launchpad") from error

    @functools.lru_cache()
    def getArchive(self):
        ppa = self.team.getPPAByName(self._ppaname)
        Logger.debug("Using PPA '%s'", ppa.web_link)
        return ppa

    def _private_ppa_url(self, filename):
        "Format the URL for a filename using the private PPA server"
        url = self.getArchive().getMySubscriptionURL()
        pkg = self.source
        subdir = pkg[:4] if pkg.startswith("lib") else pkg[:1]
        return f"{url}/pool/main/{subdir}/{pkg}/{filename}"

    def _source_urls(self, name):
        "Generator of sources for name"
        if self.getArchive().private:
            yield self._private_ppa_url(name)
        else:
            yield from super()._source_urls(name)

    def _binary_urls(self, name, bpph):
        "Generator of URLs for name"
        if self.getArchive().private:
            yield self._private_ppa_url(name)
        else:
            yield from super()._binary_urls(name, bpph)


class UbuntuCloudArchiveSourcePackage(PersonalPackageArchiveSourcePackage):
    "Download / unpack an Ubuntu Cloud Archive source package"

    TEAM = "ubuntu-cloud-archive"
    PROJECT = "cloud-archive"
    VALID_POCKETS = ["updates", "proposed", "staging"]

    def __init__(self, *args, **kwargs):
        # Need to determine actual series/pocket ppa now, as it affects getArchive()
        series, pocket = self._findReleaseAndPocketForPackage(
            kwargs.get("series"),
            kwargs.get("pocket"),
            kwargs.get("package"),
            kwargs.get("version"),
        )
        # Drop series/pocket from kwargs, as UCA handles them completely different and we
        # don't want to pass them up to the superclass
        kwargs.pop("series", None)
        orig_pocket = kwargs.pop("pocket", None)
        if orig_pocket and orig_pocket != pocket and pocket == "staging":
            Logger.info(
                "Ubuntu Cloud Archive release '%s' pocket '%s' PPA is not public,"
                " using 'staging' pocket instead",
                series,
                orig_pocket,
            )

        kwargs["ppa"] = f"{self.TEAM}/{series}-{pocket}"
        super().__init__(*args, **kwargs)

        if pocket == "staging":
            # Don't bother with the archive; just get directly from the staging ppa, since
            # none of the binaries from the archive will match the staging checksums
            self.masters = []
        else:
            self.masters = ["http://ubuntu-cloud.archive.canonical.com/ubuntu/"]

    def pull_binaries(self, arch=None, name=None, ext=None):
        """Pull binary debs into workdir.

        This is only a wrapper around the superclass method, to log warning if
        pulling binaries when using a 'staging' ppa, since the published binaries
        will not match the 'staging' binaries.
        """
        if self._ppaname.endswith("-staging"):
            Logger.warning(
                "Binaries from 'staging' pocket will not match published binaries; "
                "see https://bugs.launchpad.net/cloud-archive/+bug/1649979"
            )
        return super().pull_binaries(arch, name, ext)

    @classmethod
    @functools.lru_cache()
    def getUbuntuCloudArchiveProject(cls):
        return Project(cls.PROJECT)

    @classmethod
    def getUbuntuCloudArchiveReleaseNames(cls):
        """Get list of the main UCA release names

        The list will be sorted in descending chronological order.
        """
        return [s.name for s in cls.getUbuntuCloudArchiveProject().series]

    @classmethod
    def ppas(cls):
        """DEPRECATED: use getUbuntuCloudArchiveReleaseNames()"""
        return cls.getUbuntuCloudArchiveReleaseNames()

    @classmethod
    def isValidRelease(cls, release):
        return release in cls.getUbuntuCloudArchiveReleaseNames()

    @classmethod
    def getDevelSeries(cls):
        """Get the current UCA devel release name"""
        return cls.getUbuntuCloudArchiveReleaseNames()[0]

    @classmethod
    @functools.lru_cache()
    def getUbuntuCloudArchiveTeam(cls):
        return PersonTeam.fetch(cls.TEAM)

    @classmethod
    @functools.lru_cache()
    def getUbuntuCloudArchivePPAs(cls, release=None, pocket=None):
        """Get sorted list of UCA ppa Archive objects

        If release and/or pocket are specified, the list will be
        filtered to return only matching ppa(s).

        This list will only contain ppas relevant to UCA releases;
        it will not contain 'other' ppas, e.g. 'cloud-tools-next'.
        """
        if not any((release, pocket)):
            all_ppas = cls.getUbuntuCloudArchiveTeam().getPPAs()
            ppas = []
            for ppa_release in cls.getUbuntuCloudArchiveReleaseNames():
                for valid_pocket in cls.VALID_POCKETS:
                    name = f"{ppa_release}-{valid_pocket}"
                    if name in all_ppas:
                        ppas.append(all_ppas[name])
            return ppas

        # Use recursive call without params to get the lru-cached list of ppas returned above
        ppas = cls.getUbuntuCloudArchivePPAs()
        if release:
            ppas = list(filter(lambda p: p.name.partition("-")[0] == release, ppas))
        if pocket:
            ppas = list(filter(lambda p: p.name.partition("-")[2] == pocket, ppas))
        if not ppas:
            rname = release or "*"
            pname = pocket or "*"
            raise SeriesNotFoundException(f"UCA release '{rname}-{pname}' not found")
        return ppas

    @classmethod
    def parseReleaseAndPocket(cls, release):
        """Parse the UCA release and pocket for the given string.

        We allow 'release' to be specified as:
           1. UCARELEASE
           2. UCARELEASE-POCKET
           3. UBUNTURELEASE-UCARELEASE
           4. UBUNTURELEASE-UCARELEASE-POCKET
           5. UBUNTURELEASE-POCKET/UCARELEASE
        The UBUNTURELEASE is a standard Ubuntu release name, e.g. 'focal',
        however it is NOT checked for validity.
        The UCARELEASE is a standard Ubuntu Cloud Archive release name, e.g. 'train'.
        The POCKET is limited to the standard pockets 'release', 'updates', or 'proposed',
        or the special pocket 'staging'. The 'release' and 'updates' pockets are
        equivalent (UCA has no 'release' pocket).

        This will return a tuple of (release, pocket), or (release, None) if no
        pocket was specified.

        This will raise SeriesNotFoundException if the release and/or pocket were
        not found.
        """
        release = release.lower().strip()

        # Cases 1 and 2
        pattern1 = r"^(?P<ucarelease>[a-z]+)(?:-(?P<pocket>[a-z]+))?$"
        # Cases 3 and 4
        pattern2 = r"^(?P<ubunturelease>[a-z]+)-(?P<ucarelease>[a-z]+)(?:-(?P<pocket>[a-z]+))?$"
        # Case 5
        pattern3 = r"^(?P<ubunturelease>[a-z]+)-(?P<pocket>[a-z]+)/(?P<ucarelease>[a-z]+)$"

        for pattern in [pattern1, pattern2, pattern3]:
            match = re.match(pattern, release)
            if match:
                uca_release = match.group("ucarelease")
                pocket = match.group("pocket")
                # For UCA, there is no 'release' pocket, the default is 'updates'
                if pocket and pocket == "release":
                    Logger.warning(
                        "Ubuntu Cloud Archive does not use 'release' pocket,"
                        " using 'updates' instead"
                    )
                    pocket = "updates"
                if cls.isValidRelease(uca_release) and (not pocket or pocket in cls.VALID_POCKETS):
                    Logger.debug("Using Ubuntu Cloud Archive release '%s'", uca_release)
                    return (uca_release, pocket)
        raise SeriesNotFoundException(f"Ubuntu Cloud Archive release '{release}' not found")

    @classmethod
    def _findReleaseAndPocketForPackage(cls, release, pocket, package, version):
        if release and release not in cls.getUbuntuCloudArchiveReleaseNames():
            raise SeriesNotFoundException(f"Ubuntu Cloud Archive release '{release}' not found")
        if pocket and pocket not in cls.VALID_POCKETS:
            raise PocketDoesNotExistError(f"Ubuntu Cloud Archive pocket '{pocket}' is invalid")
        default = tuple(
            cls.getUbuntuCloudArchivePPAs(release=release or cls.getDevelSeries())[0].name.split(
                "-", maxsplit=1
            )
        )
        if not package:
            # not much we can do without a package name
            return default
        checked_pocket = False
        for ppa in cls.getUbuntuCloudArchivePPAs(release=release):
            if pocket and pocket != ppa.name.partition("-")[2]:
                # If pocket is specified, only look at the requested pocket, or 'later'
                # This allows using the 'staging' pocket for old releases that do not
                # provide any 'updates' or 'proposed' pockets
                if not checked_pocket:
                    continue
            checked_pocket = True
            params = {"exact_match": True, "source_name": package}
            if version:
                params["version"] = version
            if ppa.getPublishedSources(**params):
                ppa_release, _, ppa_pocket = ppa.name.partition("-")
                return (ppa_release, ppa_pocket)
        # package/version not found in any ppa
        return default


class _WebJSON:
    def getHostUrl(self):
        raise NotImplementedError(f"{self.__class__.__name__}.getHostUrl() is not implemented")

    def load(self, path=""):
        reader = codecs.getreader("utf-8")
        url = self.getHostUrl() + path
        Logger.debug("Loading %s", url)
        with closing(urlopen(url)) as data:
            return json.load(reader(data))


# DAKweb madison API
# https://github.com/Debian/dak/blob/master/dakweb/queries/madison.py
# This is really only useful to easily find the latest version of a
# package for a specific series (or unstable).  This does not provide
# any details at all for older-than-latest package versions.
class Madison(_WebJSON):
    urls = {
        "debian": "https://api.ftp-master.debian.org/madison",
        "ubuntu": "https://ubuntu-archive-team.ubuntu.com/madison.cgi",
    }

    def __init__(self, distro="debian"):
        super().__init__()
        self._distro = distro
        # This currently will NOT work with ubuntu; it doesn't support f=json
        if distro != "debian":
            raise InvalidDistroValueError("Madison currently only supports Debian")

    def getHostUrl(self):
        return self.urls[self._distro]

    def getSourcePackage(self, name, series="unstable"):
        url = f"?f=json&package={name}&s={series}"
        try:
            result = self.load(url)
        except HTTPError:
            result = None
        if not result:
            msg = f"Package {name} not found in '{series}'"
            raise PackageNotFoundException(msg)
        versions = list(result[0][name].values())[0]
        latest = versions[sorted(versions.keys(), reverse=True)[0]]
        return Snapshot.getSourcePackage(name=latest["source"], version=latest["source_version"])


# Snapshot API
# https://anonscm.debian.org/cgit/mirror/snapshot.debian.org.git/plain/API
class _Snapshot(_WebJSON):
    DEBIAN_COMPONENTS = ["main", "contrib", "non-free", "non-free-firmware"]

    def getHostUrl(self):
        return "http://snapshot.debian.org"

    def getComponent(self, name, version):
        # unfortunately there is no (easy) way to find the component for older
        # package versions (madison only lists the most recent versions).
        # so we have to parse the file path to determine the component :(
        url = f"/mr/package/{name}/{version}/srcfiles"
        try:
            response = self.load(f"{url}?fileinfo=1")
        except HTTPError as error:
            msg = f"Package {name} version {version} not found"
            raise PackageNotFoundException(msg) from error
        result = response.get("result")
        info = response.get("fileinfo")
        if len(result) < 1:
            msg = f"No source files for package {name} version {version}"
            raise PackageNotFoundException(msg)
        path = info[result[0]["hash"]][0]["path"]
        # this expects the 'component' to follow 'pool[-*]' in the path
        found_pool = False
        component = None
        for part in path.split("/"):
            if found_pool:
                component = part
                break
            if part.startswith("pool"):
                found_pool = True
        if not component:
            Logger.warning("could not determine component from path %s", path)
            return self.DEBIAN_COMPONENTS[0]
        if component not in self.DEBIAN_COMPONENTS:
            Logger.warning("unexpected component %s", component)
        return component

    def _get_package(self, name, url, pkginit, version, sort_key):
        try:
            results = self.load(f"/mr/{url}/{name}/")["result"]
        except HTTPError as error:
            raise PackageNotFoundException(f"Package {name} not found.") from error

        results = sorted(results, key=lambda r: r[sort_key], reverse=True)
        results = [pkginit(r) for r in results if version == r["version"]]
        if not results:
            msg = f"Package {name} version {version} not found."
            raise PackageNotFoundException(msg)
        return results

    def getSourcePackages(self, name, version):
        return self._get_package(
            name, "package", lambda obj: SnapshotSourcePackage(obj, name), version, "version"
        )

    def getSourcePackage(self, name, version):
        return self.getSourcePackages(name, version)[0]

    def getBinaryPackages(self, name, version):
        return self._get_package(name, "binary", SnapshotBinaryPackage, version, "binary_version")

    def getBinaryPackage(self, name, version):
        return self.getBinaryPackages(name, version)[0]


Snapshot = _Snapshot()


class SnapshotPackage:
    def __init__(self, obj):
        self.name = None
        self._obj = obj
        self._files = None
        self._component = None

    @property
    def version(self):
        return self._obj["version"]

    @property
    def component(self):
        if not self._component:
            self._component = Snapshot.getComponent(self.name, self.version)
        return self._component


class SnapshotSourcePackage(SnapshotPackage):
    def __init__(self, obj, name):
        # obj required fields: 'version'
        super().__init__(obj)
        self.name = name
        self._binary_files = None
        self._spph = None

    def getSPPH(self):
        if not self._spph:
            self._spph = SnapshotSPPH(self)
        return self._spph

    def getAllFiles(self):
        return self.getFiles() + self.getBinaryFiles()

    def getBinaryFiles(self, arch=None, name=None, ext=None):
        if not self._binary_files:
            url = f"/mr/package/{self.name}/{self.version}/allfiles"
            response = Snapshot.load(f"{url}?fileinfo=1")
            info = response["fileinfo"]
            files = [
                SnapshotBinaryFile(
                    b["name"],
                    b["version"],
                    self.component,
                    info[r["hash"]][0],
                    r["hash"],
                    r["architecture"],
                    self.name,
                )
                for b in response["result"]["binaries"]
                for r in b["files"]
            ]
            self._binary_files = files
        bins = list(self._binary_files)
        if arch:
            bins = [b for b in bins if b.isArch(arch)]
        if name:
            bins = [b for b in bins if re.match(name, b.package_name)]
        if ext:
            bins = [b for b in bins if re.match(ext, b.ext)]
        return bins

    def getFiles(self):
        if not self._files:
            url = f"/mr/package/{self.name}/{self.version}/srcfiles"
            response = Snapshot.load(f"{url}?fileinfo=1")
            info = response["fileinfo"]
            self._files = [
                SnapshotSourceFile(
                    self.name, self.version, self.component, info[r["hash"]][0], r["hash"]
                )
                for r in response["result"]
            ]
        return list(self._files)


class SnapshotBinaryPackage(SnapshotPackage):
    @property
    def name(self):
        return self._obj["name"]

    @property
    def binary_version(self):
        return self._obj["binary_version"]

    @property
    def source(self):
        return self._obj["source"]

    def getBPPH(self, arch):
        f = self.getFiles(arch)
        if not f:
            return None
        if not arch:
            raise RuntimeError("Must specify arch")
        # Can only be 1 binary file for this pkg name/version/arch
        return f[0].getBPPH()

    def getFiles(self, arch=None):
        if not self._files:
            url = f"/mr/binary/{self.name}/{self.version}/binfiles"
            response = Snapshot.load(f"{url}?fileinfo=1")
            info = response["fileinfo"]
            self._files = [
                SnapshotBinaryFile(
                    self.name,
                    self.version,
                    self.component,
                    info[r["hash"]][0],
                    r["hash"],
                    r["architecture"],
                    self.source,
                )
                for r in response["result"]
            ]
        if not arch:
            return list(self._files)
        return [f for f in self._files if f.isArch(arch)]


class SnapshotFile:
    def __init__(self, pkg_name, pkg_version, component, obj, h):  # pylint: disable=invalid-name
        self.package_name = pkg_name
        self.package_version = pkg_version
        self.component = component
        self._obj = obj
        self._hash = h

    @staticmethod
    def getType():
        return None

    @property
    def archive_name(self):
        return self._obj["archive_name"]

    @property
    def name(self):
        return self._obj["name"]

    @property
    def ext(self):
        return self.name.rpartition(".")[2]

    @property
    def path(self):
        return self._obj["path"]

    @property
    def size(self):
        return int(self._obj["size"])

    @property
    def date(self):
        if "run" in self._obj:
            return self._obj["run"]
        if "first_seen" in self._obj:
            return self._obj["first_seen"]
        Logger.error("File %s has no date information", self.name)
        return "unknown"

    def getHash(self):
        return self._hash

    def getUrl(self):
        return f"{Snapshot.getHostUrl()}/file/{self.getHash()}"

    def __repr__(self):
        return f"{self.path}/{self.name} {self.size} bytes {self.date}"


class SnapshotSourceFile(SnapshotFile):
    @staticmethod
    def getType():
        return "source"


class SnapshotBinaryFile(SnapshotFile):
    def __init__(self, name, version, component, obj, h, arch, source):
        super().__init__(name, version, component, obj, h)
        self.source = source
        self.arch = arch
        self._bpph = None

    def isArch(self, arch):
        if not arch:
            return True
        if self.arch == "all":
            return True
        return arch == self.arch

    @staticmethod
    def getType():
        return "binary"

    def getBPPH(self):
        if not self._bpph:
            self._bpph = SnapshotBPPH(self)
        return self._bpph


class SnapshotSPPH:
    """Provide the same interface as SourcePackagePublishingHistory"""

    def __init__(self, snapshot_pkg):
        self._pkg = snapshot_pkg
        self._changelog = None

    # LP API defined fields

    @property
    def component_name(self):
        return self.getComponent()

    @property
    def display_name(self):
        return f"{self.getPackageName()} {self.getVersion()}"

    @property
    def pocket(self):
        # Debian does not use 'pockets'
        return "Release"

    @property
    def source_package_name(self):
        return self.getPackageName()

    @property
    def source_package_version(self):
        return self.getVersion()

    # SPPH functions

    def getPackageName(self):
        return self._pkg.name

    def getVersion(self):
        return self._pkg.version

    def getComponent(self):
        return self._pkg.component

    def sourceFileUrls(self, include_meta=False):
        if include_meta:
            return [
                {
                    "url": f.getUrl(),
                    "filename": f.name,
                    "sha1": f.getHash(),
                    "sha256": None,
                    "size": f.size,
                }
                for f in self._pkg.getFiles()
            ]
        return [f.getUrl() for f in self._pkg.getFiles()]

    def sourceFileUrl(self, filename):
        for f in self.sourceFileUrls(include_meta=True):
            if filename == f["filename"]:
                return f["url"]
        return None

    def sourceFileSha1(self, url_or_filename):
        for f in self.sourceFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return f["sha1"]
        return None

    def sourceFileSha256(self, url_or_filename):  # pylint: disable=no-self-use,unused-argument
        return None

    def sourceFileSize(self, url_or_filename):
        for f in self.sourceFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return int(f["size"])
        return 0

    def getChangelog(self, since_version=None):
        """
        Return the changelog, optionally since a particular version
        May return None if the changelog isn't available
        """
        if self._changelog is None:
            name = self.getPackageName()
            if name.startswith("lib"):
                subdir = f"lib{name[3]}"
            else:
                subdir = name[0]
            pkgversion = Version(self.getVersion()).strip_epoch()
            base = "http://packages.debian.org/"

            url = os.path.join(
                base,
                "changelogs",
                "pool",
                self.getComponent(),
                subdir,
                name,
                f"{name}_{pkgversion}",
                "changelog.txt",
            )
            try:
                with closing(urlopen(url)) as f:
                    self._changelog = f.read()
            except HTTPError as error:
                Logger.error("%s: %s", url, error)
                return None

        if since_version is None:
            return self._changelog

        if isinstance(since_version, str):
            since_version = Version(since_version)

        new_entries = []
        for block in Changelog(self._changelog):
            if block.version <= since_version:
                break
            new_entries.append(str(block))
        return "".join(new_entries)

    def getBinaries(self, arch=None, name=None, ext=None):
        return [b.getBPPH() for b in self._pkg.getBinaryFiles(arch=arch, name=name, ext=ext)]


class SnapshotBPPH:  # pylint: disable=too-many-public-methods
    """Provide the same interface as BinaryPackagePublishingHistory"""

    def __init__(self, snapshot_binfile):
        self._file = snapshot_binfile

    # LP API defined fields
    @property
    def architecture_specific(self):
        return self._file.arch != "all"

    @property
    def binary_package_name(self):
        return self.getPackageName()

    @property
    def binary_package_version(self):
        return self.getVersion()

    @property
    def component_name(self):
        return self.getComponent()

    @property
    def display_name(self):
        return f"{self.getPackageName()} {self.getVersion()}"

    @property
    def pocket(self):
        # Debian does not use 'pockets'
        return "Release"

    # BPPH functions

    @property
    def arch(self):
        return self._file.arch

    def getSourcePackageName(self):
        return self._file.source

    def getPackageName(self):
        return self._file.package_name

    def getVersion(self):
        return self._file.package_version

    def getComponent(self):
        return self._file.component

    def binaryFileUrls(self, include_meta=False):
        if include_meta:
            return [
                {
                    "url": self.getUrl(),
                    "filename": self.getFileName(),
                    "sha1": self._file.getHash(),
                    "sha256": None,
                    "size": self._file.size,
                }
            ]
        return [self.getUrl()]

    def binaryFileUrl(self, filename):
        if filename == self.getFileName():
            return self.getUrl()
        return None

    def binaryFileSha1(self, url_or_filename):
        if url_or_filename in [self.getUrl(), self.getFileName()]:
            return self._file.getHash()
        return None

    def binaryFileSha256(self, url_or_filename):  # pylint: disable=no-self-use,unused-argument
        return None

    def binaryFileSize(self, url_or_filename):
        if url_or_filename in [self.getUrl(), self.getFileName()]:
            return int(self._file.size)
        return 0

    @staticmethod
    def getBuild():
        return None

    def getUrl(self):
        return self._file.getUrl()

    def getFileVersion(self):
        return self.getVersion()

    def getFileArch(self):
        return self.arch

    def getFileExt(self):
        return self._file.ext

    def getFileName(self):
        return self._file.name



# Copyright (C) 2024 Canonical Ltd.
# Author: Chris Peterson <chris.peterson@canonical.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import unittest


# Binary Tests
class BinaryTests(unittest.TestCase):

    # The requestsync binary has the option of using the launchpad api
    # to log in but requires python3-keyring in addition to
    # python3-launchpadlib. Testing the integrated login functionality
    # automatically isn't very feasbile, but we can at least write a smoke
    # test to make sure the required packages are installed.
    # See LP: #2049217
    def test_keyring_installed(self):
        """Smoke test for required lp api dependencies"""
        try:
            # pylint: disable-next=import-outside-toplevel,unused-import
            import keyring  # noqa: F401
        except ModuleNotFoundError as error:
            raise ModuleNotFoundError("package python3-keyring is not installed") from error



# -*- coding: utf-8 -*-
#
#   lpapicache.py - wrapper classes around the LP API implementing caching
#                   for usage in the ubuntu-dev-tools package
#
#   Copyright © 2009-2010 Michael Bienia <geser@ubuntu.com>
#               2011      Stefano Rivera <stefanor@ubuntu.com>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   Please see the /usr/share/common-licenses/GPL file for the full text
#   of the GNU General Public License license.
#
#   Based on code written by Jonathan Davies <jpds@ubuntu.com>

import collections
import logging
import os
import re
from copy import copy
from typing import Any
from urllib.error import URLError
from urllib.parse import urlparse

from debian.changelog import Changelog
from launchpadlib.errors import HTTPError
from launchpadlib.launchpad import Launchpad as LP
from lazr.restfulclient.resource import Entry

from ubuntutools.lp import API_VERSION, SERVICE
from ubuntutools.lp.udtexceptions import (
    AlreadyLoggedInError,
    ArchiveNotFoundException,
    ArchSeriesNotFoundException,
    PackageNotFoundException,
    PocketDoesNotExistError,
    SeriesNotFoundException,
)
from ubuntutools.misc import (
    DEFAULT_POCKETS,
    DEFAULT_STATUSES,
    POCKETS,
    STATUSES,
    download_text,
    host_architecture,
)
from ubuntutools.version import Version

Logger = logging.getLogger(__name__)


__all__ = [
    "Archive",
    "BinaryPackagePublishingHistory",
    "Build",
    "Distribution",
    "DistributionSourcePackage",
    "DistroSeries",
    "DistroArchSeries",
    "Launchpad",
    "PackageUpload",
    "PersonTeam",
    "Project",
    "ProjectSeries",
    "SourcePackagePublishingHistory",
]


class _Launchpad:
    """Singleton for LP API access."""

    __lp = None

    def login(self, service=SERVICE, api_version=API_VERSION):
        """Enforce a non-anonymous login."""
        if not self.logged_in:
            self.__lp = LP.login_with("ubuntu-dev-tools", service, version=api_version)
            # Unfortunately launchpadlib may 'login' using cached
            # credentials, without actually verifying if the credentials
            # are valid; which can lead to this 'login' not actually
            # logging in.
            # So, this forces actual LP access here, to force actual login.
            self.__lp.me  # pylint: disable=pointless-statement
        else:
            raise AlreadyLoggedInError("Already logged in to Launchpad.")

    def login_anonymously(self, service=SERVICE, api_version=API_VERSION):
        """Enforce an anonymous login."""
        if not self.logged_in:
            self.__lp = LP.login_anonymously("ubuntu-dev-tools", service, version=api_version)
        else:
            raise AlreadyLoggedInError("Already logged in to Launchpad.")

    def login_existing(self, lp):
        """Use an already logged in Launchpad object"""
        if not self.logged_in:
            self.__lp = lp
        else:
            raise AlreadyLoggedInError("Already logged in to Launchpad.")

    @property
    def logged_in(self):
        """Are we logged in?"""
        return "_Launchpad__lp" in self.__dict__

    def __getattr__(self, attr):
        if not self.logged_in:
            self.login_anonymously()
        return getattr(self.__lp, attr)

    def __call__(self):
        return self


Launchpad = _Launchpad()


class MetaWrapper(type):
    """
    A meta class used for wrapping LP API objects.
    """

    def __init__(cls, name, bases, attrd):
        super(MetaWrapper, cls).__init__(name, bases, attrd)
        if "resource_type" not in attrd:
            raise TypeError(f'Class "{name}" needs an associated resource type')
        cls._cache = {}


class BaseWrapper(metaclass=MetaWrapper):
    """
    A base class from which other wrapper classes are derived.
    """

    resource_type: str | tuple[str, str] = ""  # it's a base class after all

    def __new__(cls, data):
        if isinstance(data, str) and data.startswith(str(Launchpad._root_uri)):
            # looks like a LP API URL
            # check if it's already cached
            cached = cls._cache.get(data)
            if cached:
                return cached

            # not cached, so try to get it
            try:
                data = Launchpad.load(data)
            except HTTPError:
                # didn't work
                pass

        if isinstance(data, Entry):
            service_root, resource_type = data.resource_type_link.split("#")
            if service_root == str(Launchpad._root_uri) and resource_type in cls.resource_type:
                # check if it's already cached
                cached = cls._cache.get(data.self_link)
                if not cached:
                    # create a new instance
                    cached = object.__new__(cls)
                    cached._lpobject = data
                    # and add it to our cache
                    cls._cache[data.self_link] = cached
                    Logger.debug("%s: %s", cls.__name__, data.self_link)
                    # add additional class specific caching (if available)
                    cache = getattr(cls, "cache", None)
                    if isinstance(cache, collections.abc.Callable):
                        cache(cached)
                return cached
            raise TypeError(f"'{data}' is not a '{cls.resource_type}' object")

        # not a LP API representation, let the specific class handle it
        fetch = getattr(cls, "fetch", None)
        if isinstance(fetch, collections.abc.Callable):
            return fetch(data)
        raise NotImplementedError(f"Don't know how to fetch '{data}' from LP")

    def __call__(self):
        return self._lpobject

    def __getattr__(self, attr):
        return getattr(self._lpobject, attr)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._lpobject!r}>"


class Distribution(BaseWrapper):
    """
    Wrapper class around a LP distribution object.
    """

    resource_type = "distribution"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._archives = {}
        self._series_by_name = {}
        self._series = {}
        self._dev_series = None
        self._have_all_series = False
        self._main_archive = None

    def cache(self):
        self._cache[self.name] = self

    def _cache_series(self, series):
        """
        Add the DistroSeries to the cache if needed.
        """
        if series.version not in self._series:
            self._series_by_name[series.name] = series
            self._series[series.version] = series

    @classmethod
    def fetch(cls, dist):
        """
        Fetch the distribution object identified by 'dist' from LP.
        """
        if not isinstance(dist, str):
            raise TypeError(f"Don't know what do with '{dist!r}'")
        cached = cls._cache.get(dist)
        if not cached:
            cached = Distribution(Launchpad.distributions[dist])
        return cached

    def getArchive(self, archive=None):
        """
        Returns an Archive object for the requested archive.
        Raises a ArchiveNotFoundException if the archive doesn't exist.

        If 'archive' is None, return the main archive.
        """
        if archive:
            res = self._archives.get(archive)

            if not res:
                for archive_ in self.archives:
                    if archive_.name == archive:
                        res = Archive(archive_)
                        self._archives[res.name] = res
                        break

            if res:
                return res
            message = f"The Archive '{archive}' doesn't exist in {self.display_name}"
            raise ArchiveNotFoundException(message)

        if self._main_archive is None:
            self._main_archive = Archive(self.main_archive_link)
        return self._main_archive

    def getSeries(self, name_or_version):
        """
        Returns a DistroSeries object for a series passed by name
        (e.g. 'karmic') or version (e.g. '9.10').
        If the series is not found: raise SeriesNotFoundException
        """
        if name_or_version in self._series:
            return self._series[name_or_version]
        if name_or_version in self._series_by_name:
            return self._series_by_name[name_or_version]

        try:
            series = DistroSeries(self().getSeries(name_or_version=name_or_version))
        except HTTPError as error:
            message = f"Release '{name_or_version}' is unknown in '{self.display_name}'."
            raise SeriesNotFoundException(message) from error

        self._cache_series(series)
        return series

    def getDevelopmentSeries(self):
        """
        Returns a DistroSeries object of the current development series.
        """
        if not self._dev_series:
            series = DistroSeries(self.current_series_link)
            self._cache_series(series)
            self._dev_series = series
        return self._dev_series

    def getAllSeries(self, active=True):  # pylint: disable=unused-argument
        """
        Returns a list of all DistroSeries objects.
        """
        if not self._have_all_series:
            for series in self.series:
                self._cache_series(DistroSeries(series))
            self._have_all_series = True

        allseries = filter(lambda s: s.active, self._series.values())
        allseries = sorted(allseries, key=lambda s: float(s.version), reverse=True)
        Logger.debug("Found series: %s", ", ".join([f"{s.name} ({s.version})" for s in allseries]))
        return collections.OrderedDict((s.name, s) for s in allseries)


class DistroArchSeries(BaseWrapper):  # pylint: disable=too-few-public-methods
    """
    Wrapper class around a LP distro arch series object.
    """

    resource_type = "distro_arch_series"

    def getSeries(self):
        """
        Get DistroSeries for this.
        """
        return DistroSeries(self._lpobject.distroseries_link)


class DistroSeries(BaseWrapper):
    """
    Wrapper class around a LP distro series object.
    """

    resource_type = "distro_series"

    def __init__(self, *args):  # pylint: disable=unused-argument
        if "_architectures" not in self.__dict__:
            self._architectures = {}

    def getArchSeries(self, archtag=None):
        """
        Returns a DistroArchSeries object for an architecture passed by name
        (e.g. 'amd64').
        If arch is not specified, get the DistroArchSeries for the system arch.
        The special archtag 'all' will get the system arch.
        If the architecture is not found: raise ArchSeriesNotFoundException.
        """
        if not archtag or archtag == "all":
            archtag = host_architecture()
        if archtag not in self._architectures:
            try:
                architecture = DistroArchSeries(self().getDistroArchSeries(archtag=archtag))
                self._architectures[architecture.architecture_tag] = architecture
            except HTTPError as error:
                message = f"Architecture {archtag} is unknown."
                raise ArchSeriesNotFoundException(message) from error
        return self._architectures[archtag]

    def getPackageUploads(self, name=None, pocket=None, version=None, status="Unapproved"):
        """Returns a list of PackageUploads for this series."""
        params = {"exact_match": True}
        if name:
            params["name"] = name
        if pocket:
            params["pocket"] = pocket
        if version:
            params["version"] = version
        if status:
            params["status"] = status
        return [PackageUpload(p) for p in self._lpobject.getPackageUploads(**params)]


class PackageUpload(BaseWrapper):
    """
    Wrapper class around a LP package_upload object.
    """

    resource_type = "package_upload"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._custom_urls = None
        self._source_urls = None
        self._binary_urls = None
        self._binary_properties = None
        self._binary_prop_dict = None

    def getArchive(self):
        return Archive(self._lpobject.archive_link)

    def getSourceArchive(self):
        if self._lpobject.copy_source_archive_link:
            return Archive(self._lpobject.copy_source_archive_link)
        return None

    def getDistroSeries(self):
        return DistroSeries(self._lpobject.distroseries_link)

    def changesFileUrl(self):
        return self._lpobject.changes_file_url

    def customFileUrls(self):
        if not self._custom_urls:
            self._custom_urls = self._lpobject.customFileUrls()
        return copy(self._custom_urls)

    def sourceFileUrls(self):
        if not self._source_urls:
            self._source_urls = self._lpobject.sourceFileUrls()
        return copy(self._source_urls)

    def binaryFileUrls(self):
        if not self._binary_urls:
            self._binary_urls = self._lpobject.binaryFileUrls()
        return copy(self._binary_urls)

    def getBinaryProperties(self):
        if not self._binary_properties:
            self._binary_properties = self._lpobject.getBinaryProperties()
        return copy(self._binary_properties)

    def binaryFileProperties(self, filename_or_url):
        if not self._binary_prop_dict:
            urls = self.binaryFileUrls()
            props = self.getBinaryProperties()
            self._binary_prop_dict = dict(zip(urls, props))
            for key, value in copy(self._binary_prop_dict).items():
                filename = os.path.basename(urlparse(key).path)
                self._binary_prop_dict[filename] = value
        return self._binary_prop_dict.get(filename_or_url, {})


class Archive(BaseWrapper):
    """
    Wrapper class around a LP archive object.
    """

    resource_type = "archive"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._binpkgs = {}
        self._srcpkgs = {}
        self._pkg_uploaders = {}
        self._pkgset_uploaders = {}
        self._component_uploaders = {}

    def getSourcePackage(
        self,
        name,
        series=None,
        pocket=None,
        version=None,
        status=None,
        wrapper=None,
        search_all_series=False,
    ):
        """
        Returns a SourcePackagePublishingHistory object for the most
        recent source package in the distribution 'dist', series and
        pocket.

        series defaults to the current development series if not specified.
        series must be either a series name string, or DistroSeries object.

        version may be specified to get only the exact version requested.

        pocket may be a string or a list.  If no version is provided, it
        defaults to all pockets except 'Backports'; if searching for a
        specific version, it defaults to all pockets.  Pocket strings must
        be capitalized.

        status may be a string or a list.  If no version is provided, it
        defaults to only 'Pending' and 'Published'; if searching for a
        specific version, it defaults to all statuses.  Status strings must
        be capitalized.

        wrapper is the class to return an instance of; defaults to
        SourcePackagePublishingHistory.

        search_all_series is used if series is None.  If False, this will
        search only the latest devel series, and if True all series
        will be searched, in reverse order, starting with the latest
        devel series.  Defaults to False.

        If the requested source package doesn't exist a
        PackageNotFoundException is raised.
        """
        return self._getPublishedItem(
            name,
            series,
            pocket,
            cache=self._srcpkgs,
            function="getPublishedSources",
            name_key="source_name",
            wrapper=wrapper or SourcePackagePublishingHistory,
            version=version,
            status=status,
            search_all_series=search_all_series,
            binary=False,
        )

    def getBinaryPackage(
        self,
        name,
        archtag=None,
        series=None,
        pocket=None,
        version=None,
        status=None,
        wrapper=None,
        search_all_series=False,
    ):
        """
        Returns a BinaryPackagePublishingHistory object for the most
        recent source package in the distribution 'dist', architecture
        'archtag', series and pocket.

        series defaults to the current development series if not specified.
        series must be either a series name string, or DistroArchSeries object.
        series may be omitted if version is specified.

        version may be specified to get only the exact version requested.

        pocket may be a string or a list.  If no version is provided, it
        defaults to all pockets except 'Backports'; if searching for a
        specific version, it defaults to all pockets.  Pocket strings must
        be capitalized.

        status may be a string or a list.  If no version is provided, it
        defaults to only 'Pending' and 'Published'; if searching for a
        specific version, it defaults to all statuses.  Status strings must
        be capitalized.

        wrapper is the class to return an instance of; defaults to
        BinaryPackagePublishingHistory.

        search_all_series is used if series is None.  If False, this will
        search only the latest devel series, and if True all series
        will be searched, in reverse order, starting with the latest
        devel series.  Defaults to False.

        If the requested binary package doesn't exist a
        PackageNotFoundException is raised.
        """
        return self._getPublishedItem(
            name,
            series,
            pocket,
            archtag=archtag,
            cache=self._binpkgs,
            function="getPublishedBinaries",
            name_key="binary_name",
            wrapper=wrapper or BinaryPackagePublishingHistory,
            version=version,
            status=status,
            search_all_series=search_all_series,
            binary=True,
        )

    def _getPublishedItem(
        self,
        name,
        series,
        pocket,
        cache,
        function,
        name_key,
        wrapper,
        archtag=None,
        version=None,
        status=None,
        search_all_series=False,
        binary=False,
    ):
        """
        Common code between getSourcePackage and getBinaryPackage.

        Don't use this directly.
        """
        if not pocket:
            if version and not series:
                # check ALL pockets if specific version in any series
                pockets = POCKETS
            else:
                # otherwise, check all pockets EXCEPT 'Backports'
                pockets = DEFAULT_POCKETS
        elif isinstance(pocket, str):
            pockets = (pocket,)
        else:
            pockets = tuple(pocket)

        for pocket_ in pockets:
            if pocket_ not in POCKETS:
                raise PocketDoesNotExistError(f"Pocket '{pocket_}' does not exist.")

        if not status:
            if version:
                # check ALL statuses if specific version
                statuses = STATUSES
            else:
                # otherwise, only check 'Pending' and 'Published'
                statuses = DEFAULT_STATUSES
        elif isinstance(status, str):
            statuses = (status,)
        else:
            statuses = tuple(status)

        for status_ in statuses:
            if status_ not in STATUSES:
                raise ValueError(f"Status '{status_}' is not valid.")

        dist = Distribution(self.distribution_link)

        # please don't pass DistroArchSeries as archtag!
        # but, the code was like that before so keep
        # backwards compatibility.
        if isinstance(archtag, DistroArchSeries):
            series = archtag
            archtag = None

        series_to_check = [series]
        if not version and not series:
            # if neither version or series are specified, use either the
            # devel series or search all series
            if search_all_series:
                series_to_check = dist.getAllSeries().values()
            else:
                series_to_check = [dist.getDevelopmentSeries()]

        # check each series - if only version was provided, series will be None
        for series in series_to_check:  # pylint: disable=redefined-argument-from-local
            arch_series = None

            if isinstance(series, DistroArchSeries):
                arch_series = series
                series = series.getSeries()
            elif isinstance(series, DistroSeries):
                pass
            elif series:
                series = dist.getSeries(series)

            if binary:
                if arch_series is None and series:
                    arch_series = series.getArchSeries(archtag=archtag)
                if archtag is None and arch_series:
                    archtag = arch_series.architecture_tag
                if archtag is None:
                    archtag = host_architecture()

            index = (name, getattr(series, "name", None), archtag, pockets, statuses, version)

            if index in cache:
                return cache[index]

            params = {name_key: name, "exact_match": True}

            if arch_series:
                params["distro_arch_series"] = arch_series()
            elif series:
                params["distro_series"] = series()

            if len(pockets) == 1:
                params["pocket"] = pockets[0]

            if len(statuses) == 1:
                params["status"] = statuses[0]

            if version:
                params["version"] = version

            Logger.debug(
                "Calling %s(%s)", function, ", ".join([f"{k}={v}" for (k, v) in params.items()])
            )
            records = getattr(self, function)(**params)

            err_msg = f"does not exist in the {dist.display_name} {self.name} archive"

            for record in records:
                if binary:
                    rversion = getattr(record, "binary_package_version", None)
                else:
                    rversion = getattr(record, "source_package_version", None)

                if record.pocket not in pockets:
                    err_msg = f"pocket {record.pocket} not in ({','.join(pockets)})"
                    Logger.debug("Skipping version %s: %s", rversion, err_msg)
                    continue
                if record.status not in statuses:
                    err_msg = f"status {record.status} not in ({','.join(statuses)})"
                    Logger.debug("Skipping version %s: %s", rversion, err_msg)
                    continue
                release = wrapper(record)
                if binary and archtag and archtag != release.arch:
                    err_msg = f"arch {release.arch} does not match requested arch {archtag}"
                    Logger.debug("Skipping version %s: %s", rversion, err_msg)
                    continue
                # results are ordered so first is latest
                cache[index] = release
                return release

        version_with_epoch = None
        if version and version == Version(version).strip_epoch() and len(records) == 0:
            # a specific version was asked for, but we found none;
            # check if one exists with an epoch to give a hint in error msg
            for epoch in range(1, 9):
                version_ = Version(version)
                version_.epoch = epoch
                params["version"] = version_.full_version
                if len(getattr(self, function)(**params)) > 0:
                    version_with_epoch = version_.full_version
                    Logger.debug("Found version with epoch %s", version_with_epoch)
                    break

        if name_key == "binary_name":
            package_type = "binary package"
        elif name_key == "source_name":
            package_type = "source package"
        else:
            package_type = "package"
        msg = f"The {package_type} '{name}' "
        if version:
            msg += f"version {version} "
        msg += err_msg
        if binary and archtag:
            msg += f" for architecture {archtag}"
        if len(series_to_check) > 1:
            msg += " in any release"
            if len(pockets) == 1:
                msg += f" for pocket {pockets[0]}"
            elif len(pockets) != len(POCKETS):
                msg += f" for pockets {', '.join(pockets)}"
        elif series:
            msg += f" in {series.name}"
            if len(pockets) == 1:
                msg += f"-{pockets[0]}"
            elif len(pockets) != len(POCKETS):
                msg += f" for pockets {', '.join(pockets)}"
        if len(statuses) == 1:
            msg += f" with status {statuses[0]}"
        elif len(statuses) != len(STATUSES):
            msg += f" with status in {', '.join(statuses)}"
        if version_with_epoch:
            msg += f" (did you forget the epoch? try {version_with_epoch})"
        raise PackageNotFoundException(msg)

    def copyPackage(
        self,
        source_name,
        version,
        from_archive,
        to_pocket,
        to_series=None,
        sponsored=None,
        include_binaries=False,
    ):
        """Copy a single named source into this archive.

        Asynchronously copy a specific version of a named source to the
        destination archive if necessary.  Calls to this method will return
        immediately if the copy passes basic security checks and the copy
        will happen sometime later with full checking.
        """
        # pylint: disable=protected-access
        if isinstance(sponsored, PersonTeam):
            sponsored = sponsored._lpobject

        self._lpobject.copyPackage(
            source_name=source_name,
            version=version,
            from_archive=from_archive._lpobject,
            to_pocket=to_pocket,
            to_series=to_series,
            sponsored=sponsored,
            include_binaries=include_binaries,
        )

    def getUploadersForComponent(self, component_name):
        """Get the list of PersonTeams who can upload packages in the
        specified component.
        [Note: the permission records, themselves, aren't exposed]
        """
        if component_name not in self._component_uploaders:
            self._component_uploaders[component_name] = sorted(
                set(
                    PersonTeam(permission.person_link)
                    for permission in self._lpobject.getUploadersForComponent(
                        component_name=component_name
                    )
                )
            )
        return self._component_uploaders[component_name]

    def getUploadersForPackage(self, source_package_name):
        """Get the list of PersonTeams who can upload source_package_name)
        [Note: the permission records, themselves, aren't exposed]
        """
        if source_package_name not in self._pkg_uploaders:
            self._pkg_uploaders[source_package_name] = sorted(
                set(
                    PersonTeam(permission.person_link)
                    for permission in self._lpobject.getUploadersForPackage(
                        source_package_name=source_package_name
                    )
                ),
                key=lambda s: s.name,
            )
        return self._pkg_uploaders[source_package_name]

    def getUploadersForPackageset(self, packageset, direct_permissions=False):
        """Get the list of PersonTeams who can upload packages in packageset
        [Note: the permission records, themselves, aren't exposed]
        """
        key = (packageset, direct_permissions)
        if key not in self._pkgset_uploaders:
            self._pkgset_uploaders[key] = sorted(
                set(
                    PersonTeam(permission.person_link)
                    # pylint: disable=protected-access
                    for permission in self._lpobject.getUploadersForPackageset(
                        packageset=packageset._lpobject, direct_permissions=direct_permissions
                    )
                )
            )
        return self._pkgset_uploaders[key]

    def getMySubscriptionURL(self):
        """Get the "subscription URL" for the logged in user

        If this is a private archive (i.e. private PPA), this returns
        the "subscription URL" including authentication; otherwise
        this returns None.
        """
        if self.private:
            return PersonTeam.me.getArchiveSubscriptionURL(archive=self._lpobject)
        return None


class SourcePackagePublishingHistory(BaseWrapper):
    """
    Wrapper class around a LP source package object.
    """

    resource_type = "source_package_publishing_history"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._archive = None
        self._changelog = None
        self._binaries = {}
        self._distro_series = None
        self._source_urls = None
        # Don't share _builds between different
        # SourcePackagePublishingHistory objects
        if "_builds" not in self.__dict__:
            self._builds = {}

    def getDistroSeries(self):
        """
        Return the DistroSeries.
        """
        if not self._distro_series:
            self._distro_series = DistroSeries(self._lpobject.distro_series_link)
        return self._distro_series

    def getPackageName(self):
        """
        Returns the source package name.
        """
        return self._lpobject.source_package_name

    def getVersion(self):
        """
        Returns the version of the source package.
        """
        return self._lpobject.source_package_version

    def getComponent(self):
        """
        Returns the component of the source package.
        """
        return self._lpobject.component_name

    def getSeriesName(self):
        """
        Returns the series

        Named getSeriesName() to avoid confusion with
        getDistroSeries()
        """
        return self.getDistroSeries().name

    def getSeriesAndPocket(self):
        """
        Returns a human-readable release-pocket
        """
        release = self.getSeriesName()
        if self.pocket != "Release":
            release += f"-{self.pocket.lower()}"
        return release

    def getArchive(self):
        """
        Get this SPPH's archive.
        """
        if not self._archive:
            self._archive = Archive(self._lpobject.archive_link)

        return self._archive

    def getChangelog(self, since_version=None):
        """
        Return the changelog, optionally since a particular version
        May return None if the changelog isn't available
        Only available in the devel API, not 1.0
        """
        if self._changelog is None:
            url = self._lpobject.changelogUrl()
            if url is None:
                Logger.error(
                    "No changelog available for %s %s", self.getPackageName(), self.getVersion()
                )
                return None

            try:
                self._changelog = download_text(url)
            except URLError as e:
                Logger.error("Exception while downloading '%s': %s", url, e)
                return None

        if since_version is None:
            return self._changelog

        if isinstance(since_version, str):
            since_version = Version(since_version)

        new_entries = []
        for block in Changelog(self._changelog):
            if block.version <= since_version:
                break
            new_entries.append(str(block))
        return "".join(new_entries)

    def sourceFileUrls(self, include_meta=False):
        """
        Return the URL for this source publication's files.

        The include_meta param changes the return value;
        when it is False (the default), an array of url strings is
        returned.  When include_meta is True, an array is returned
        with dicts, containing the entries:
          url: the url string
          sha1: the SHA1 checksum of the source file (if provided)
          sha256: the SHA256 checksum of the source file
          size: the size of the source file
        Also, this function adds a 'filename' field:
          filename: the filename parsed from the url path
        Note that while all the keys will be in the dict, their values
        may be None.
        """
        if not self._source_urls:
            urls = self._lpobject.sourceFileUrls(include_meta=True)
            if not urls:
                Logger.warning(
                    "SPPH %s_%s has no sourceFileUrls", self.getPackageName(), self.getVersion()
                )
            for url in urls:
                # make sure mandatory fields are present
                for field in ["url", "sha1", "sha256", "size"]:
                    if field not in url:
                        url[field] = None
                url["filename"] = os.path.basename(urlparse(url["url"]).path)
            self._source_urls = urls

        if include_meta:
            return list(self._source_urls)
        return [f["url"] for f in self._source_urls]

    def sourceFileUrl(self, filename):
        """
        Returns the URL for the specified source filename.

        If the filename is not found in the sourceFileUrls(), this returns None.
        """
        for f in self.sourceFileUrls(include_meta=True):
            if filename == f["filename"]:
                return f["url"]
        return None

    def sourceFileSha1(self, url_or_filename):
        """
        Returns the SHA1 checksum for the specified source file url.

        If the url is not found in the sourceFileUrls(), this returns None.

        The url may be specified as a filename.
        """
        for f in self.sourceFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return f["sha1"]
        return None

    def sourceFileSha256(self, url_or_filename):
        """
        Returns the SHA256 checksum for the specified source file url.

        If the url is not found in the sourceFileUrls(), this returns None.

        The url may be specified as a filename.
        """
        for f in self.sourceFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return f["sha256"]
        return None

    def sourceFileSize(self, url_or_filename):
        """
        Returns the size for the specified source file url.

        If the url is not found in the sourceFileUrls(), this returns 0.

        The url may be specified as a filename.
        """
        for f in self.sourceFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return int(f["size"])
        return 0

    def getBinaries(self, arch=None, name=None, ext=None):
        """
        Returns the resulting BinaryPackagePublishingHistorys.
        If arch is specified, it returns binaries for only that arch,
        plus any binaries with arch 'all'.  If arch is not specified, or
        if arch is specified as 'all', all archs are returned.

        If name is specified, only returns BPPH matching that (regex) name.

        If ext is specified, only returns BPPH matching that (regex) ext.
        """
        if arch == "all":
            arch = None

        if self.status in ["Pending", "Published"]:
            # Published, great!  Directly query the list of binaries
            binaries = map(BinaryPackagePublishingHistory, self._lpobject.getPublishedBinaries())
            for binary in binaries:
                arch_ = binary.arch
                if arch_ not in self._binaries:
                    self._binaries[arch_] = {}
                self._binaries[arch_][binary.binary_package_name] = binary
        else:
            # we have to go the long way :(
            Logger.info("Please wait, this may take some time...")
            archive = self.getArchive()
            urls = self.binaryFileUrls()
            for url in urls:
                # strip out the URL leading text.
                filename = os.path.basename(urlparse(url).path)
                # strip the file suffix
                pkgname, _, extension = filename.rpartition(".")
                # split into name, version, arch
                name_, _, arch_ = pkgname.rsplit("_", 2)
                # arch 'all' has separate bpph for each real arch,
                # but all point to the same binary url
                if arch_ == "all":
                    arch_ = arch or host_architecture()
                # Only check the arch requested - saves time
                if arch and arch != arch_:
                    continue
                # Only check the name requested - saves time
                if name and not re.match(name, name_):
                    continue
                # Only check the ext requested - saves time
                if ext and not re.match(ext, extension):
                    continue
                # If we already have this BPPH, keep going
                if arch_ in self._binaries and name_ in self._binaries[arch_]:
                    continue
                # we ignore the version, as it may be missing epoch
                # also we can't use series, as some package versions
                # span multiple series! (e.g. for different archs)
                params = {"name": name_, "archtag": arch_, "version": self.getVersion()}
                try:
                    bpph = archive.getBinaryPackage(**params)
                except PackageNotFoundException:
                    Logger.debug("Could not find pkg in archive: %s", filename)
                    continue
                if arch_ not in self._binaries:
                    self._binaries[arch_] = {}
                self._binaries[arch_][name_] = bpph

        if not arch:
            bpphs = [b for a in self._binaries.values() for b in a.values()]
        elif arch in self._binaries:
            bpphs = list(self._binaries[arch].values())
        else:
            return []

        if name:
            bpphs = [b for b in bpphs if re.match(name, b.binary_package_name)]

        if ext:
            bpphs = [b for b in bpphs if re.match(ext, b.getFileExt())]

        return bpphs

    def _fetch_builds(self):
        """Populate self._builds with the build records."""
        builds = self.getBuilds()
        for build in builds:
            self._builds[build.arch_tag] = Build(build)


class BinaryPackagePublishingHistory(BaseWrapper):
    """
    Wrapper class around a LP binary package object.
    """

    resource_type = "binary_package_publishing_history"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._arch = None
        self._ext = None
        self._binary_urls = None

    @property
    def arch(self):
        if not self._arch:
            das = DistroArchSeries(self._lpobject.distro_arch_series_link)
            self._arch = das.architecture_tag
        return self._arch

    def getSourcePackageName(self):
        """
        Returns the source package name.
        """
        return self.getBuild().source_package_name

    def getPackageName(self):
        """
        Returns the binary package name.
        """
        return self._lpobject.binary_package_name

    def getVersion(self):
        """
        Returns the version of the binary package.
        """
        return self._lpobject.binary_package_version

    def getComponent(self):
        """
        Returns the component of the binary package.
        """
        return self._lpobject.component_name

    def binaryFileUrls(self, include_meta=False):
        """
        Return the URL for this binary publication's files.
        Only available in the devel API, not 1.0

        The include_meta param changes the return value;
        when it is False (the default), an array of url strings is
        returned (but typically there is only a single url in the array).
        When include_meta is True, an array (again, with typically only one
        entry) is returned with dicts, containing the entries:
          url: the url string
          sha1: the SHA1 checksum of the binary file
          sha256: the SHA256 checksum of the binary file
          size: the size of the binary file
        Also, this function adds a 'filename' field:
          filename: the filename parsed from the url path
        Note that while all the keys will be in the dict, their values
        may be None.
        """
        if not self._binary_urls:
            try:
                urls = self._lpobject.binaryFileUrls(include_meta=True)
            except AttributeError as error:
                raise AttributeError(
                    "binaryFileUrls can only be found in lpapi "
                    "devel, not 1.0. Login using devel to have it."
                ) from error
            if not urls:
                Logger.warning(
                    "BPPH %s_%s has no binaryFileUrls", self.getPackageName(), self.getVersion()
                )
            for url in urls:
                # make sure mandatory fields are present
                for field in ["url", "sha1", "sha256", "size"]:
                    if field not in url:
                        url[field] = None
                url["filename"] = os.path.basename(urlparse(url["url"]).path)
            self._binary_urls = urls

        if include_meta:
            return list(self._binary_urls)
        return [f["url"] for f in self._binary_urls]

    def binaryFileUrl(self, filename):
        """
        Returns the URL for the specified binary filename.

        If the filename is not found in the binaryFileUrls(), this returns None.
        """
        for f in self.binaryFileUrls(include_meta=True):
            if filename == f["filename"]:
                return f["url"]
        return None

    def binaryFileSha1(self, url_or_filename):
        """
        Returns the SHA1 checksum for the specified binary file url.

        If the url is not found in the binaryFileUrls(), this returns None.

        The url may be specified as a filename.
        """
        for f in self.binaryFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return f["sha1"]
        return None

    def binaryFileSha256(self, url_or_filename):
        """
        Returns the SHA256 checksum for the specified binary file url.

        If the url is not found in the binaryFileUrls(), this returns None.

        The url may be specified as a filename.
        """
        for f in self.binaryFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return f["sha256"]
        return None

    def binaryFileSize(self, url_or_filename):
        """
        Returns the size for the specified binary file url.

        If the url is not found in the binaryFileUrls(), this returns 0.

        The url may be specified as a filename.
        """
        for f in self.binaryFileUrls(include_meta=True):
            if url_or_filename in [f["url"], f["filename"]]:
                return int(f["size"])
        return 0

    def getBuild(self):
        """
        Returns the original build of the binary package.
        """
        return Build(self._lpobject.build_link)

    def getUrl(self):
        """
        Returns the original build URL of the binary package.
        """
        return f"{self.getBuild().getUrl()}/+files/{self.getFileName()}"

    def getFileVersion(self):
        """
        Returns the file version, which is the package version without the epoch
        """
        return Version(self.getVersion()).strip_epoch()

    def getFileArch(self):
        """
        Returns the file arch, which is 'all' if not arch-specific
        """
        if bool(self._lpobject.architecture_specific):
            return self.arch
        return "all"

    def getFileExt(self):
        """
        Returns the file extension; "deb", "ddeb", or "udeb".
        """
        if not self._ext:
            self._ext = self._getFileExt()

        return self._ext

    def _getFileExt(self):
        try:
            # this is the best way, from the actual URL filename
            return self.binaryFileUrls()[0].rpartition(".")[2]
        except (AttributeError, IndexError):
            Logger.debug("Could not get file ext from url, trying to guess...")

        # is_debug should be reliable way of detecting ddeb...?
        try:
            if self.is_debug:
                return "ddeb"
        except AttributeError:
            # is_debug only available with api version 'devel'
            if self.getPackageName().endswith("-dbgsym"):
                return "ddeb"

        # is this reliable?
        if self.getPackageName().endswith("-di") or self.getPackageName().endswith("-udeb"):
            return "udeb"

        # everything else - assume regular deb
        return "deb"

    def getFileName(self):
        """
        Returns the filename for this binary package.
        """
        return (
            f"{self.getPackageName()}_{self.getFileVersion()}"
            f"_{self.getFileArch()}.{self.getFileExt()}"
        )


class MetaPersonTeam(MetaWrapper):
    @property
    def me(cls):
        """The PersonTeam object of the currently authenticated LP user or
        None when anonymously logged in.
        """
        if "_me" not in cls.__dict__:
            try:
                # We have to use me.self_link due to LP: #504297
                cls._me = PersonTeam(Launchpad.me.self_link)
            except HTTPError as error:
                if error.response.status == 401:
                    # Anonymous login
                    cls._me = None
                else:
                    raise
        return cls._me


class PersonTeam(BaseWrapper, metaclass=MetaPersonTeam):
    """
    Wrapper class around a LP person or team object.
    """

    resource_type = ("person", "team")

    def __init__(self, *args):  # pylint: disable=unused-argument
        # Don't share _upload between different PersonTeams
        self._ppas = None
        if "_upload" not in self.__dict__:
            self._upload = {}

    def __str__(self):
        return f"{self.display_name} ({self.name})"

    def cache(self):
        self._cache[self.name] = self

    @classmethod
    def fetch(cls, person_or_team):
        """
        Fetch the person or team object identified by 'url' from LP.
        """
        if not isinstance(person_or_team, str):
            raise TypeError(f"Don't know what do with '{person_or_team!r}'")
        cached = cls._cache.get(person_or_team)
        if not cached:
            cached = PersonTeam(Launchpad.people[person_or_team])
        return cached

    def isLpTeamMember(self, team):
        """
        Checks if the user is a member of a certain team on Launchpad.

        Returns True if the user is a member of the team otherwise False.
        """
        return any(t.name == team for t in self.super_teams)

    def canUploadPackage(self, archive, distroseries, package, component, pocket="Release"):
        """Check if the person or team has upload rights for the source
        package to the specified 'archive' and 'distrorelease'.

        A source package name and a component have to be specified.
        'archive' has to be a Archive object.
        'distroseries' has to be an DistroSeries object.
        """
        if not isinstance(archive, Archive):
            raise TypeError(f"'{archive!r}' is not an Archive object.")
        if not isinstance(distroseries, DistroSeries):
            raise TypeError(f"'{distroseries!r}' is not a DistroSeries object.")
        if package is not None and not isinstance(package, str):
            raise TypeError("A source package name expected.")
        if component is not None and not isinstance(component, str):
            raise TypeError("A component name expected.")
        if package is None and component is None:
            raise ValueError("Either a source package name or a component has to be specified.")
        if pocket not in POCKETS:
            raise PocketDoesNotExistError(f"Pocket '{pocket}' does not exist.")

        can_upload = self._upload.get((archive, distroseries, pocket, package, component))

        if can_upload is None:
            # checkUpload() throws an exception if the person can't upload
            try:
                archive.checkUpload(
                    component=component,
                    distroseries=distroseries(),
                    person=self(),
                    pocket=pocket,
                    sourcepackagename=package,
                )
                can_upload = True
            except HTTPError as e:
                if e.response.status == 403:
                    can_upload = False
                else:
                    raise e
            index = (archive, distroseries, pocket, package, component)
            self._upload[index] = can_upload

        return can_upload

    def getPPAs(self):
        if self._ppas is None:
            ppas = [Archive(ppa) for ppa in self._lpobject.ppas]
            self._ppas = {ppa.name: ppa for ppa in ppas}
        return self._ppas

    def getPPAByName(self, name):
        return Archive(self._lpobject.getPPAByName(name=name))


class Project(BaseWrapper):
    """
    Wrapper class around a LP project object.
    """

    resource_type = "project"

    def __init__(self, *args):  # pylint: disable=unused-argument
        self._series = None

    @property
    def series(self):
        """Get a list of all ProjectSeries

        The list will be sorted by date_created, in descending order.
        """
        if not self._series:
            series = [ProjectSeries(s) for s in self._lpobject.series]
            self._series = sorted(series, key=lambda s: s.date_created, reverse=True)
        return self._series.copy()

    @classmethod
    def fetch(cls, project):
        """
        Fetch the project object identified by 'project' from LP.
        """
        if not isinstance(project, str):
            raise TypeError(f"Don't know what do with '{project!r}'")
        return Project(Launchpad.projects(project))


class ProjectSeries(BaseWrapper):  # pylint: disable=too-few-public-methods
    """
    Wrapper class around a LP project_series object.
    """

    resource_type = "project_series"


class Build(BaseWrapper):
    """
    Wrapper class around a build object.
    """

    resource_type = "build"

    def __str__(self):
        return f"{self.arch_tag}: {self.buildstate}"

    def getSourcePackagePublishingHistory(self):
        link = self._lpobject.current_source_publication_link
        if link:
            if re.search("redacted", link):
                # Too old - the link has been 'redacted'
                return None
            return SourcePackagePublishingHistory(link)
        return None

    def getUrl(self):
        return self()

    def rescore(self, score):
        if self.can_be_rescored:
            self().rescore(score=score)
            return True
        return False

    def retry(self):
        if self.can_be_retried:
            self().retry()
            return True
        return False


class DistributionSourcePackage(BaseWrapper):  # pylint: disable=too-few-public-methods
    """
    Caching class for distribution_source_package objects.
    """

    resource_type = "distribution_source_package"


class Packageset(BaseWrapper):  # pylint: disable=too-few-public-methods
    """
    Caching class for packageset objects.
    """

    resource_type = "packageset"
    _lp_packagesets = None
    _source_sets: dict[tuple[str, str | None, bool], Any] = {}

    @classmethod
    def setsIncludingSource(cls, sourcepackagename, distroseries=None, direct_inclusion=False):
        """Get the package sets including sourcepackagename"""

        if cls._lp_packagesets is None:
            cls._lp_packagesets = Launchpad.packagesets

        key = (sourcepackagename, distroseries, direct_inclusion)
        if key not in cls._source_sets:
            params = {"sourcepackagename": sourcepackagename, "direct_inclusion": direct_inclusion}
            if distroseries is not None:
                params["distroseries"] = distroseries._lpobject  # pylint: disable=protected-access

            cls._source_sets[key] = [
                Packageset(packageset)
                for packageset in cls._lp_packagesets.setsIncludingSource(**params)
            ]

        return cls._source_sets[key]



# pullpkg.py -- pull package files for debian/ubuntu/uca
#               modified from ../pull-lp-source and converted to module
#
# Copyright (C) 2008,      Iain Lane <iain@orangesquash.org.uk>,
#               2010-2011, Stefano Rivera <stefanor@ubuntu.com>
#               2017,      Dan Streetman <dan.streetman@canonical.com>
#
# ##################################################################
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# See file /usr/share/common-licenses/GPL for more details.
#
# ##################################################################


import errno
import logging
import os
import re
import subprocess
import sys
from argparse import ArgumentParser
from urllib.parse import urlparse

from distro_info import DebianDistroInfo

# by default we use standard logging.getLogger() and only use
# ubuntutools.getLogger() in PullPkg().main()
from ubuntutools import getLogger as ubuntutools_getLogger
from ubuntutools.archive import (
    DebianSourcePackage,
    PersonalPackageArchiveSourcePackage,
    UbuntuCloudArchiveSourcePackage,
    UbuntuSourcePackage,
)
from ubuntutools.config import UDTConfig
from ubuntutools.lp.lpapicache import Distribution, Launchpad
from ubuntutools.lp.udtexceptions import (
    AlreadyLoggedInError,
    InvalidDistroValueError,
    PackageNotFoundException,
    PocketDoesNotExistError,
    SeriesNotFoundException,
)
from ubuntutools.misc import (
    STATUSES,
    UPLOAD_QUEUE_STATUSES,
    download,
    host_architecture,
    split_release_pocket,
)

Logger = logging.getLogger(__name__)

PULL_SOURCE = "source"
PULL_DEBS = "debs"
PULL_DDEBS = "ddebs"
PULL_UDEBS = "udebs"
PULL_LIST = "list"

VALID_PULLS = [PULL_SOURCE, PULL_DEBS, PULL_DDEBS, PULL_UDEBS, PULL_LIST]
VALID_BINARY_PULLS = [PULL_DEBS, PULL_DDEBS, PULL_UDEBS]

DISTRO_DEBIAN = "debian"
DISTRO_UBUNTU = "ubuntu"
DISTRO_UCA = "uca"
DISTRO_PPA = "ppa"

DISTRO_PKG_CLASS = {
    DISTRO_DEBIAN: DebianSourcePackage,
    DISTRO_UBUNTU: UbuntuSourcePackage,
    DISTRO_UCA: UbuntuCloudArchiveSourcePackage,
    DISTRO_PPA: PersonalPackageArchiveSourcePackage,
}
VALID_DISTROS = DISTRO_PKG_CLASS.keys()


class InvalidPullValueError(ValueError):
    """Thrown when --pull value is invalid"""


class PullPkg:
    """Class used to pull file(s) associated with a specific package"""

    @classmethod
    def main(cls, *args, **kwargs):
        """For use by stand-alone cmdline scripts.

        This will handle catching certain exceptions or kbd interrupts,
        setting up the root logger level to INFO, and printing out
        (via Logger) a caught error message, instead of allowing the
        exception to flow up to the script.  This does not catch
        unexpected exceptions, such as internal errors.

        On (expected) error, this will call sys.exit(error);
        unexpected errors will flow up to the caller.
        On success, this simply returns.
        """
        logger = ubuntutools_getLogger()

        try:
            cls(*args, **kwargs).pull()
            return
        except KeyboardInterrupt:
            logger.info("User abort.")
        except (
            PackageNotFoundException,
            SeriesNotFoundException,
            PocketDoesNotExistError,
            InvalidDistroValueError,
            InvalidPullValueError,
        ) as error:
            logger.error(str(error))
            sys.exit(errno.ENOENT)

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        self._default_pull = kwargs.get("pull")
        self._default_distro = kwargs.get("distro")
        self._default_arch = kwargs.get("arch", host_architecture())

    def parse_args(self, args):
        if args is None:
            args = sys.argv[1:]

        help_default_pull = "What to pull: " + ", ".join(VALID_PULLS)
        if self._default_pull:
            help_default_pull += f" (default: {self._default_pull})"
        help_default_distro = "Pull from: " + ", ".join(VALID_DISTROS)
        if self._default_distro:
            help_default_distro += f" (default: {self._default_distro})"
        help_default_arch = "Get binary packages for arch"
        help_default_arch += f"(default: {self._default_arch})"

        # use add_help=False because we do parse_known_args() below, and if
        # that sees --help then it exits immediately
        parser = ArgumentParser(add_help=False)
        parser.add_argument("-L", "--login", action="store_true", help="Login to Launchpad")
        parser.add_argument(
            "-v", "--verbose", action="count", default=0, help="Increase verbosity/debug"
        )
        parser.add_argument(
            "-d", "--download-only", action="store_true", help="Do not extract the source package"
        )
        parser.add_argument("-m", "--mirror", action="append", help="Preferred mirror(s)")
        parser.add_argument(
            "--no-conf",
            action="store_true",
            help="Don't read config files or environment variables",
        )
        parser.add_argument(
            "--no-verify-signature",
            action="store_true",
            help="Don't fail if dsc signature can't be verified",
        )
        parser.add_argument(
            "-s",
            "--status",
            action="append",
            default=[],
            help="Search for packages with specific status(es)",
        )
        parser.add_argument("-a", "--arch", default=self._default_arch, help=help_default_arch)
        parser.add_argument("-p", "--pull", default=self._default_pull, help=help_default_pull)
        parser.add_argument(
            "-D", "--distro", default=self._default_distro, help=help_default_distro
        )

        # add distro-specific params
        try:
            distro = self.parse_distro(parser.parse_known_args(args)[0].distro)
        except InvalidDistroValueError:
            # don't fail at this point, finish setting up parser help/usage
            distro = None

        if distro == DISTRO_UBUNTU:
            parser.add_argument(
                "--security",
                action="store_true",
                help="Pull from the Ubuntu Security Team (proposed) PPA",
            )
            parser.add_argument(
                "--upload-queue", action="store_true", help="Pull from the Ubuntu upload queue"
            )
        if distro == DISTRO_PPA:
            parser.add_argument("--ppa", help="PPA to pull from")
            if parser.parse_known_args(args)[0].ppa is None:
                # check for any param starting with "ppa:"
                # if found, move it to a --ppa param
                for param in args:
                    if param.startswith("ppa:"):
                        args.remove(param)
                        args.insert(0, param)
                        args.insert(0, "--ppa")
                        break

        # add the positional params
        parser.add_argument("package", help="Package name to pull")
        parser.add_argument("release", nargs="?", help="Release to pull from")
        parser.add_argument("version", nargs="?", help="Package version to pull")

        epilog = (
            "Note on --status: if a version is provided, all status types "
            "will be searched; if no version is provided, by default only "
            "'Pending' and 'Published' status will be searched."
        )

        # since parser has no --help handler, create a new parser that does
        newparser = ArgumentParser(parents=[parser], epilog=epilog)

        return self.parse_options(vars(newparser.parse_args(args)))

    @staticmethod
    def parse_pull(pull):
        if not pull:
            raise InvalidPullValueError("Must specify --pull")

        # allow 'dbgsym' as alias for 'ddebs'
        if pull == "dbgsym":
            Logger.debug("Pulling '%s' for '%s'", PULL_DDEBS, pull)
            pull = PULL_DDEBS
        # assume anything starting with 'bin' means 'debs'
        if str(pull).startswith("bin"):
            Logger.debug("Pulling '%s' for '%s'", PULL_DEBS, pull)
            pull = PULL_DEBS
        # verify pull action is valid
        if pull not in VALID_PULLS:
            raise InvalidPullValueError(f"Invalid pull action '{pull}'")

        return pull

    @staticmethod
    def parse_distro(distro):
        if not distro:
            raise InvalidDistroValueError("Must specify --distro")

        distro = distro.lower()

        # allow 'lp' for 'ubuntu'
        if distro == "lp":
            Logger.debug("Using distro '%s' for '%s'", DISTRO_UBUNTU, distro)
            distro = DISTRO_UBUNTU
        # assume anything with 'cloud' is UCA
        if re.match(r".*cloud.*", distro):
            Logger.debug("Using distro '%s' for '%s'", DISTRO_UCA, distro)
            distro = DISTRO_UCA
        # verify distro is valid
        if distro not in VALID_DISTROS:
            raise InvalidDistroValueError(f"Invalid distro '{distro}'")

        return distro

    @staticmethod
    def parse_release(distro, release):
        if distro == DISTRO_UCA:
            return UbuntuCloudArchiveSourcePackage.parseReleaseAndPocket(release)

        # Check if release[-pocket] is specified
        release, pocket = split_release_pocket(release, default=None)
        Logger.debug("Parsed release '%s' pocket '%s'", release, pocket)

        if distro == DISTRO_DEBIAN:
            # This converts from the aliases like 'unstable'
            debian_info = DebianDistroInfo()
            codename = debian_info.codename(release)
            if codename:
                Logger.info("Using release '%s' for '%s'", codename, release)
                release = codename

        if distro == DISTRO_PPA:
            # PPAs are part of Ubuntu distribution
            distribution = Distribution(DISTRO_UBUNTU)
        else:
            distribution = Distribution(distro)

        # let SeriesNotFoundException flow up
        distribution.getSeries(release)

        Logger.debug("Using distro '%s' release '%s' pocket '%s'", distro, release, pocket)
        return (release, pocket)

    def parse_release_and_version(self, distro, release, version, try_swap=True):
        # Verify specified release is valid, and params in correct order
        pocket = None
        try:
            release, pocket = self.parse_release(distro, release)
        except (SeriesNotFoundException, PocketDoesNotExistError):
            if try_swap:
                Logger.debug("Param '%s' not valid series, must be version", release)
                release, version = version, release
                if release:
                    return self.parse_release_and_version(distro, release, version, False)
            else:
                Logger.error("Can't find series for '%s' or '%s'", release, version)
                raise
        return (release, version, pocket)

    def parse_options(self, options):
        # if any of these fail, there is a problem with the parser
        # they should all be provided, though the optional ones may be None

        # type bool
        assert "verbose" in options
        assert "download_only" in options
        assert "no_conf" in options
        assert "no_verify_signature" in options
        assert "status" in options
        # type string
        assert "pull" in options
        assert "distro" in options
        assert "arch" in options
        assert "package" in options
        # type string, optional
        assert "release" in options
        assert "version" in options
        # type list of strings, optional
        assert "mirror" in options

        options["pull"] = self.parse_pull(options["pull"])
        options["distro"] = self.parse_distro(options["distro"])

        # ensure these are always included so we can just check for None/False later
        options["ppa"] = options.get("ppa", None)
        options["security"] = options.get("security", False)
        options["upload_queue"] = options.get("upload_queue", False)

        return options

    def _get_params(self, options):
        distro = options["distro"]
        pull = options["pull"]

        params = {}
        params["package"] = options["package"]
        params["arch"] = options["arch"]

        if options["release"]:
            release, version, pocket = self.parse_release_and_version(
                distro, options["release"], options["version"]
            )
            params["series"] = release
            params["version"] = version
            params["pocket"] = pocket

        if params["package"].endswith(".dsc") and not params["series"] and not params["version"]:
            params["dscfile"] = params["package"]
            params.pop("package")

        if options["security"]:
            if options["ppa"]:
                Logger.warning("Both --security and --ppa specified, ignoring --ppa")
            Logger.debug("Checking Ubuntu Security PPA")
            # --security is just a shortcut for --ppa ppa:ubuntu-security-proposed/ppa
            options["ppa"] = "ubuntu-security-proposed/ppa"

        if options["ppa"]:
            if options["ppa"].startswith("ppa:"):
                params["ppa"] = options["ppa"][4:]
            else:
                params["ppa"] = options["ppa"]
        elif distro == DISTRO_PPA:
            raise ValueError("Must specify PPA to pull from")

        mirrors = []
        if options["mirror"]:
            mirrors.extend(options["mirror"])
        if pull == PULL_DDEBS:
            config = UDTConfig(options["no_conf"])
            ddebs_mirror = config.get_value(distro.upper() + "_DDEBS_MIRROR")
            if ddebs_mirror:
                mirrors.append(ddebs_mirror)
        if mirrors:
            Logger.debug("using mirrors %s", ", ".join(mirrors))
            params["mirrors"] = mirrors

        params["verify_signature"] = not options["no_verify_signature"]

        params["status"] = STATUSES if "all" in options["status"] else options["status"]

        # special handling for upload queue
        if options["upload_queue"]:
            if len(options["status"]) > 1:
                raise ValueError(
                    "Too many --status provided, can only search for a single status or 'all'"
                )
            if not options["status"]:
                params["status"] = None
            elif options["status"][0].lower() == "all":
                params["status"] = "all"
            elif options["status"][0].capitalize() in UPLOAD_QUEUE_STATUSES:
                params["status"] = options["status"][0].capitalize()
            else:
                raise ValueError(
                    f"Invalid upload queue status '{options['status'][0]}':"
                    f" valid values are {', '.join(UPLOAD_QUEUE_STATUSES)}"
                )

        return params

    def pull(self, args=None):
        """Pull (download) specified package file(s)"""
        options = self.parse_args(args)

        if options["verbose"]:
            Logger.setLevel(logging.DEBUG)
            if options["verbose"] > 1:
                logging.getLogger(__package__).setLevel(logging.DEBUG)

        Logger.debug("pullpkg options: %s", options)

        pull = options["pull"]
        distro = options["distro"]

        if options["login"]:
            Logger.debug("Logging in to Launchpad:")
            try:
                Launchpad.login()
            except AlreadyLoggedInError:
                Logger.error(
                    "Launchpad singleton has already performed a login, "
                    "and its design prevents another login"
                )
                Logger.warning("Continuing anyway, with existing Launchpad instance")

        params = self._get_params(options)
        package = params["package"]

        if options["upload_queue"]:
            # upload queue API is different/simpler
            self.pull_upload_queue(  # pylint: disable=missing-kwoa
                pull, download_only=options["download_only"], **params
            )
            return

        # call implementation, and allow exceptions to flow up to caller
        srcpkg = DISTRO_PKG_CLASS[distro](**params)
        spph = srcpkg.lp_spph

        Logger.info("Found %s", spph.display_name)

        # The VCS detection logic was modeled after `apt source`
        for key in srcpkg.dsc.keys():
            original_key = key
            key = key.lower()

            if key.startswith("vcs-"):
                if key == "vcs-browser":
                    continue
                if key == "vcs-git":
                    vcs = "Git"
                elif key == "vcs-bzr":
                    vcs = "Bazaar"
                else:
                    continue

                uri = srcpkg.dsc[original_key]

                Logger.warning(
                    "\nNOTICE: '%s' packaging is maintained in "
                    "the '%s' version control system at:\n %s\n",
                    package,
                    vcs,
                    uri,
                )

                vcscmd = ""
                if vcs == "Bazaar":
                    vcscmd = " $ bzr branch " + uri
                elif vcs == "Git":
                    vcscmd = " $ git clone " + uri

                if vcscmd:
                    Logger.info(
                        "Please use:\n%s\n"
                        "to retrieve the latest (possibly unreleased) updates to the package.\n",
                        vcscmd,
                    )

        if pull == PULL_LIST:
            Logger.info("Source files:")
            for f in srcpkg.dsc["Files"]:
                Logger.info("  %s", f["name"])
            Logger.info("Binary files:")
            for f in spph.getBinaries(options["arch"]):
                archtext = ""
                name = f.getFileName()
                if name.rpartition(".")[0].endswith("all"):
                    archtext = f" ({f.arch})"
                Logger.info("  %s%s", name, archtext)
        elif pull == PULL_SOURCE:
            # allow DownloadError to flow up to caller
            srcpkg.pull()
            if options["download_only"]:
                Logger.debug("--download-only specified, not extracting")
            else:
                srcpkg.unpack()
        elif pull in VALID_BINARY_PULLS:
            name = None
            if package != spph.getPackageName():
                Logger.info("Pulling only binary package '%s'", package)
                Logger.info(
                    "Use package name '%s' to pull all binary packages", spph.getPackageName()
                )
                name = package

            # e.g. 'debs' -> 'deb'
            ext = pull.rstrip("s")

            if distro == DISTRO_DEBIAN:
                # Debian ddebs don't use .ddeb extension, unfortunately :(
                if pull in [PULL_DEBS, PULL_DDEBS]:
                    name = name or ".*"
                    ext = "deb"
                if pull == PULL_DEBS:
                    name += r"(?<!-dbgsym)$"
                if pull == PULL_DDEBS:
                    name += r"-dbgsym$"

            # allow DownloadError to flow up to caller
            total = srcpkg.pull_binaries(name=name, ext=ext, arch=options["arch"])
            if total < 1:
                Logger.error("No %s found for %s %s", pull, package, spph.getVersion())
        else:
            Logger.error("Internal error: invalid pull value after parse_pull()")
            raise InvalidPullValueError(f"Invalid pull value '{pull}'")

    def pull_upload_queue(
        self,
        pull,
        *,
        package,
        version=None,
        arch=None,
        series=None,
        pocket=None,
        status=None,
        download_only=None,
        **kwargs,
    ):  # pylint: disable=no-self-use,unused-argument
        if not series:
            Logger.error("Using --upload-queue requires specifying series")
            return

        series = Distribution("ubuntu").getSeries(series)

        queueparams = {"name": package}
        if pocket:
            queueparams["pocket"] = pocket

        if status == "all":
            queueparams["status"] = None
            queuetype = "any"
        elif status:
            queueparams["status"] = status
            queuetype = status
        else:
            queuetype = "Unapproved"

        packages = [
            p
            for p in series.getPackageUploads(**queueparams)
            if p.package_version == version or str(p.id) == version or not version
        ]

        if pull == PULL_SOURCE:
            packages = [p for p in packages if p.contains_source]
        elif pull in VALID_BINARY_PULLS:
            packages = [
                p
                for p in packages
                if p.contains_build
                and (arch in ["all", "any"] or arch in p.display_arches.replace(",", "").split())
            ]

        if not packages:
            msg = f"Package {package} not found in {queuetype} upload queue for {series.name}"
            if version:
                msg += f" with version/id {version}"
            if pull in VALID_BINARY_PULLS:
                msg += f" for arch {arch}"
            raise PackageNotFoundException(msg)

        if pull == PULL_LIST:
            for pkg in packages:
                msg = f"Found {pkg.package_name} {pkg.package_version} (ID {pkg.id})"
                if pkg.display_arches:
                    msg += f" arch {pkg.display_arches}"
                Logger.info(msg)
                url = pkg.changesFileUrl()
                if url:
                    Logger.info("Changes file:")
                    Logger.info("  %s", url)
                else:
                    Logger.info("No changes file")
                urls = pkg.sourceFileUrls()
                if urls:
                    Logger.info("Source files:")
                    for url in urls:
                        Logger.info("  %s", url)
                else:
                    Logger.info("No source files")
                urls = pkg.binaryFileUrls()
                if urls:
                    Logger.info("Binary files:")
                    for url in urls:
                        Logger.info("  %s", url)
                        Logger.info("    { %s }", pkg.binaryFileProperties(url))
                else:
                    Logger.info("No binary files")
                urls = pkg.customFileUrls()
                if urls:
                    Logger.info("Custom files:")
                    for url in urls:
                        Logger.info("  %s", url)
            return

        if len(packages) > 1:
            msg = "Found multiple packages"
            if version:
                msg += f" with version {version}, please specify the ID instead"
            else:
                msg += ", please specify the version"
            Logger.error("Available package versions/ids are:")
            for pkg in packages:
                Logger.error("%s %s (id %s)", pkg.package_name, pkg.package_version, pkg.id)
            raise PackageNotFoundException(msg)

        pkg = packages[0]

        urls = set(pkg.customFileUrls())
        if pkg.changesFileUrl():
            urls.add(pkg.changesFileUrl())

        if pull == PULL_SOURCE:
            urls |= set(pkg.sourceFileUrls())
            if not urls:
                Logger.error("No source files to download")
            dscfile = None
            for url in urls:
                dst = download(url, os.getcwd())
                if dst.name.endswith(".dsc"):
                    dscfile = dst
            if download_only:
                Logger.debug("--download-only specified, not extracting")
            elif not dscfile:
                Logger.error("No source dsc file found, cannot extract")
            else:
                cmd = ["dpkg-source", "-x", dscfile.name]
                Logger.debug(" ".join(cmd))
                result = subprocess.run(
                    cmd,
                    check=False,
                    encoding="utf-8",
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
                if result.returncode != 0:
                    Logger.error("Source unpack failed.")
                    Logger.debug(result.stdout)
        else:
            name = ".*"
            if pull == PULL_DEBS:
                name = rf"{name}(?<!-di)(?<!-dbgsym)$"
            elif pull == PULL_DDEBS:
                name += "-dbgsym$"
            elif pull == PULL_UDEBS:
                name += "-di$"
            else:
                raise InvalidPullValueError(f"Invalid pull value {pull}")

            urls |= set(pkg.binaryFileUrls())
            if not urls:
                Logger.error("No binary files to download")
            for url in urls:
                filename = os.path.basename(urlparse(url).path)
                if re.match(name, filename):
                    download(url, os.getcwd())



# test_update_maintainer.py - Test suite for ubuntutools.update_maintainer
#
# Copyright (C) 2010, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Test suite for ubuntutools.update_maintainer"""

import os
import unittest
from io import StringIO
from unittest import mock

from ubuntutools.update_maintainer import update_maintainer

_LUCID_CHANGELOG = """\
axis2c (1.6.0-0ubuntu8) lucid; urgency=low

  * rebuild rest of main for armel armv7/thumb2 optimization;
    UbuntuSpec:mobile-lucid-arm-gcc-v7-thumb2

 -- Alexander Sack <asac@ubuntu.com>  Fri, 05 Mar 2010 03:10:28 +0100
"""

_AXIS2C_CONTROL = """\
Source: axis2c
Section: libs
Priority: optional
DM-Upload-Allowed: yes
XSBC-Original-Maintainer: Soren Hansen <soren@ubuntu.com>
Maintainer: Kyo Lee <kyo.lee@eucalyptus.com>
Standards-Version: 3.9.1
Homepage: http://ws.apache.org/axis2/c/

Package: libaxis2c0
Architecture: any
Depends: ${shlibs:Depends}, ${misc:Depends}
"""

_AXIS2C_UPDATED = """\
Source: axis2c
Section: libs
Priority: optional
DM-Upload-Allowed: yes
XSBC-Original-Maintainer: Kyo Lee <kyo.lee@eucalyptus.com>
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
Standards-Version: 3.9.1
Homepage: http://ws.apache.org/axis2/c/

Package: libaxis2c0
Architecture: any
Depends: ${shlibs:Depends}, ${misc:Depends}
"""

_UNSTABLE_CHANGELOG = """\
adblock-plus (1.3.3-1) unstable; urgency=low

  * New upstream release.

 -- Benjamin Drung <bdrung@ubuntu.com>  Sat, 25 Dec 2010 20:17:41 +0100
"""

_ABP_CONTROL = """\
Source: adblock-plus
Section: web
Priority: optional
Maintainer: Dmitry E. Oboukhov <unera@debian.org>
Uploaders: Debian Mozilla Extension Maintainers <pkg-mozext-maintainers@lists.alioth.debian.org>,
           Benjamin Drung <bdrung@ubuntu.com>
Build-Depends: debhelper (>= 7.0.50~), mozilla-devscripts (>= 0.19~)
Standards-Version: 3.9.1
DM-Upload-Allowed: yes
Homepage: http://adblockplus.org/
VCS-Browser: http://git.debian.org/?p=pkg-mozext/adblock-plus.git;a=summary
VCS-Git: git://git.debian.org/pkg-mozext/adblock-plus.git

Package: xul-ext-adblock-plus
"""

_ABP_UPDATED = """\
Source: adblock-plus
Section: web
Priority: optional
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
XSBC-Original-Maintainer: Dmitry E. Oboukhov <unera@debian.org>
Uploaders: Debian Mozilla Extension Maintainers <pkg-mozext-maintainers@lists.alioth.debian.org>,
           Benjamin Drung <bdrung@ubuntu.com>
Build-Depends: debhelper (>= 7.0.50~), mozilla-devscripts (>= 0.19~)
Standards-Version: 3.9.1
DM-Upload-Allowed: yes
Homepage: http://adblockplus.org/
VCS-Browser: http://git.debian.org/?p=pkg-mozext/adblock-plus.git;a=summary
VCS-Git: git://git.debian.org/pkg-mozext/adblock-plus.git

Package: xul-ext-adblock-plus
"""

_ABP_OLD_MAINTAINER = """\
Source: adblock-plus
Section: web
Priority: optional
Maintainer: Ubuntu MOTU Developers <ubuntu-motu@lists.ubuntu.com>
XSBC-Original-Maintainer: Dmitry E. Oboukhov <unera@debian.org>
Uploaders: Debian Mozilla Extension Maintainers <pkg-mozext-maintainers@lists.alioth.debian.org>,
           Benjamin Drung <bdrung@ubuntu.com>
Build-Depends: debhelper (>= 7.0.50~), mozilla-devscripts (>= 0.19~)
Standards-Version: 3.9.1
DM-Upload-Allowed: yes
Homepage: http://adblockplus.org/
VCS-Browser: http://git.debian.org/?p=pkg-mozext/adblock-plus.git;a=summary
VCS-Git: git://git.debian.org/pkg-mozext/adblock-plus.git

Package: xul-ext-adblock-plus
"""

_SIMPLE_RULES = """\
#!/usr/bin/make -f
%:
\tdh $@
"""

_COMPLEX_RULES = """\
#!/usr/bin/make -f
... from python2.6 ...
distribution := $(shell lsb_release -is)

control-file:
\tsed -e "s/@PVER@/$(PVER)/g" \\
\t\t-e "s/@VER@/$(VER)/g" \\
\t\t-e "s/@PYSTDDEP@/$(PYSTDDEP)/g" \\
\t\t-e "s/@PRIO@/$(PY_PRIO)/g" \\
\t\t-e "s/@MINPRIO@/$(PY_MINPRIO)/g" \\
\t\tdebian/control.in > debian/control.tmp
ifeq ($(distribution),Ubuntu)
  ifneq (,$(findstring ubuntu, $(PKGVERSION)))
\tm='Ubuntu Core Developers <ubuntu-devel-discuss@lists.ubuntu.com>'; \\
\t\tsed -i "/^Maintainer:/s/\\(.*\\)/Maintainer: $$m\nXSBC-Original-\1/" \\
\t\tdebian/control.tmp
  endif
endif
\t[ -e debian/control ] \\
\t\t&& cmp -s debian/control debian/control.tmp \\
\t\t&& rm -f debian/control.tmp && exit 0; \\
\t\tmv debian/control.tmp debian/control
... from python2.6 ...
"""

_SEAHORSE_PLUGINS_CONTROL = """\
# This file is autogenerated. DO NOT EDIT!
#
# Modifications should be made to debian/control.in instead.
# This file is regenerated automatically in the clean target.

Source: seahorse-plugins
Section: gnome
Priority: optional
Maintainer: Emilio Pozuelo Monfort <pochu@debian.org>
Build-Depends: debhelper (>= 5)
Standards-Version: 3.8.3
Homepage: http://live.gnome.org/Seahorse

Package: seahorse-plugins
"""

_SEAHORSE_PLUGINS_UPDATED = """\
# This file is autogenerated. DO NOT EDIT!
#
# Modifications should be made to debian/control.in instead.
# This file is regenerated automatically in the clean target.

Source: seahorse-plugins
Section: gnome
Priority: optional
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
XSBC-Original-Maintainer: Emilio Pozuelo Monfort <pochu@debian.org>
Build-Depends: debhelper (>= 5)
Standards-Version: 3.8.3
Homepage: http://live.gnome.org/Seahorse

Package: seahorse-plugins
"""


# pylint: disable=R0904
class UpdateMaintainerTestCase(unittest.TestCase):
    """TestCase object for ubuntutools.update_maintainer"""

    _directory = "/"
    _files = {"changelog": None, "control": None, "control.in": None, "rules": None}

    def _fake_isfile(self, filename):
        """Check only for existing fake files."""
        directory, base = os.path.split(filename)
        return (
            directory == self._directory and base in self._files and self._files[base] is not None
        )

    def _fake_open(self, filename, mode="r", encoding=None):
        """Provide StringIO objects instead of real files."""
        self.assertTrue(encoding, f"encoding for {filename} not specified")
        directory, base = os.path.split(filename)
        if (
            directory != self._directory
            or base not in self._files
            or (mode == "r" and self._files[base] is None)
        ):
            raise IOError(f"No such file or directory: '{filename}'")
        if mode == "w":
            self._files[base] = StringIO()
            self._files[base].close = lambda: None
        return self._files[base]

    # pylint: disable=C0103
    def setUp(self):
        m = mock.mock_open()
        m.side_effect = self._fake_open
        patcher = mock.patch("builtins.open", m)
        self.addCleanup(patcher.stop)
        patcher.start()
        m = mock.MagicMock(side_effect=self._fake_isfile)
        patcher = mock.patch("os.path.isfile", m)
        self.addCleanup(patcher.stop)
        patcher.start()
        self._files["rules"] = StringIO(_SIMPLE_RULES)
        # Logger.stdout = StringIO()
        # Logger.stderr = StringIO()

    def tearDown(self):
        # self.assertEqual(Logger.stdout.getvalue(), '')
        # self.assertEqual(Logger.stderr.getvalue(), '')
        self._files["changelog"] = None
        self._files["control"] = None
        self._files["control.in"] = None
        self._files["rules"] = None
        # Logger.stdout = sys.stdout
        # Logger.stderr = sys.stderr

    # pylint: enable=C0103
    def test_debian_package(self):
        """Test: Don't update Maintainer field if target is Debian."""
        self._files["changelog"] = StringIO(_UNSTABLE_CHANGELOG)
        self._files["control"] = StringIO(_ABP_CONTROL)
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _ABP_CONTROL)

    def test_original_ubuntu_maintainer(self):
        """Test: Original maintainer is Ubuntu developer.

        The Maintainer field needs to be update even if
        XSBC-Original-Maintainer has an @ubuntu.com address."""
        self._files["changelog"] = StringIO(_LUCID_CHANGELOG)
        self._files["control"] = StringIO(_AXIS2C_CONTROL)
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _AXIS2C_UPDATED)
        # warnings = Logger.stderr.getvalue().strip()
        # Logger.stderr = StringIO()
        # self.assertEqual(len(warnings.splitlines()), 1)
        # self.assertRegex(warnings, "Warning: Overwriting original maintainer: "
        # "Soren Hansen <soren@ubuntu.com>")

    def test_update_maintainer(self):
        """Test: Update Maintainer field."""
        self._files["changelog"] = StringIO(_LUCID_CHANGELOG)
        self._files["control"] = StringIO(_ABP_CONTROL)
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _ABP_UPDATED)

    def test_update_old_maintainer(self):
        """Test: Update old MOTU address."""
        self._files["changelog"] = StringIO(_UNSTABLE_CHANGELOG)
        self._files["control.in"] = StringIO(_ABP_OLD_MAINTAINER)
        update_maintainer(self._directory, True)
        self.assertEqual(self._files["control.in"].getvalue(), _ABP_UPDATED)

    def test_comments_in_control(self):
        """Test: Update Maintainer field in a control file containing
        comments."""
        self._files["changelog"] = StringIO(_LUCID_CHANGELOG)
        self._files["control"] = StringIO(_SEAHORSE_PLUGINS_CONTROL)
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _SEAHORSE_PLUGINS_UPDATED)

    def test_skip_smart_rules(self):
        """Test: Skip update when XSBC-Original in debian/rules."""
        self._files["changelog"] = StringIO(_LUCID_CHANGELOG)
        self._files["control"] = StringIO(_ABP_CONTROL)
        self._files["rules"] = StringIO(_COMPLEX_RULES)
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _ABP_CONTROL)

    def test_missing_rules(self):
        """Test: Skip XSBC-Original test when debian/rules is missing."""
        self._files["changelog"] = StringIO(_LUCID_CHANGELOG)
        self._files["control"] = StringIO(_ABP_CONTROL)
        self._files["rules"] = None
        update_maintainer(self._directory)
        self.assertEqual(self._files["control"].getvalue(), _ABP_UPDATED)
